import os
from typing import List
from unittest import mock

import pytest
from click.testing import CliRunner, Result
from ruamel.yaml import YAML

from great_expectations import DataContext
from great_expectations import __version__ as ge_version
from great_expectations.cli import cli
from tests.cli.utils import assert_no_logging_messages_or_tracebacks

yaml = YAML()
yaml.default_flow_style = False


@pytest.mark.parametrize("invocation", [None, "--help", "--v3-api --help"])
def test_cli_command_help(caplog, invocation):
    runner = CliRunner(mix_stderr=True)
    result = runner.invoke(cli, invocation, catch_exceptions=False)
    assert result.exit_code == 0
    # Note that click 8.0 fixed a longstanding bug in 7.x which added empty
    # lines in "Options". This test looks for chunks of output to support both.
    assert (
        """Usage: great_expectations [OPTIONS] COMMAND [ARGS]...

  Welcome to the great_expectations CLI!

  Most commands follow this format: great_expectations <NOUN> <VERB>

  The nouns are: checkpoint, datasource, docs, init, project, store, suite,
  validation-operator. Most nouns accept the following verbs: new, list, edit"""
        in result.output
    )
    assert (
        """Options:
  --version                Show the version and exit.
  --v3-api / --v2-api      Default to v3 (Batch Request) API. Use --v2-api for
                           v2 (Batch Kwargs) API"""
        in result.output
    )
    assert (
        "  -v, --verbose            Set great_expectations to use verbose output."
        in result.output
    )
    assert (
        """  -c, --config TEXT        Path to great_expectations configuration file
                           location (great_expectations.yml). Inferred if not
                           provided."""
        in result.output
    )
    assert (
        """  -y, --assume-yes, --yes  Assume "yes" for all prompts.
  --help                   Show this message and exit.
"""
        in result.output
    )
    assert (
        """
Commands:
  checkpoint  Checkpoint operations
  datasource  Datasource operations
  docs        Data Docs operations
  init        Initialize a new Great Expectations project.
  project     Project operations
  store       Store operations
  suite       Expectation Suite operations
"""
        in result.output
    )
    assert_no_logging_messages_or_tracebacks(caplog, result)


def test_cli_command_invalid_command(caplog):
    runner = CliRunner(mix_stderr=False)
    result = runner.invoke(cli, "--v3-api blarg")
    assert result.exit_code == 2
    assert "Error: No such command" in result.stderr
    assert ("'blarg'" in result.stderr) or ('"blarg"' in result.stderr)


def test_cli_ge_version_exists(caplog):
    runner = CliRunner(mix_stderr=True)
    result = runner.invoke(cli, "--v3-api --version", catch_exceptions=False)
    assert ge_version in str(result.output)
    assert_no_logging_messages_or_tracebacks(caplog, result)


def test_cli_works_from_adjacent_directory_without_config_flag(
    monkeypatch, empty_data_context
):
    """We don't care about the NOUN here just combinations of the config flag"""
    runner = CliRunner(mix_stderr=True)
    monkeypatch.chdir(os.path.dirname(empty_data_context.root_directory))
    result = runner.invoke(cli, "--v3-api checkpoint list", catch_exceptions=False)
    assert result.exit_code == 0
    assert "No Checkpoints found" in result.output


def test_cli_works_from_great_expectations_directory_without_config_flag(
    monkeypatch, empty_data_context
):
    """We don't care about the NOUN here just combinations of the config flag"""
    runner = CliRunner(mix_stderr=True)
    monkeypatch.chdir(empty_data_context.root_directory)
    result = runner.invoke(cli, "--v3-api checkpoint list", catch_exceptions=False)
    assert result.exit_code == 0
    assert "No Checkpoints found" in result.output


def test_cli_works_from_random_directory_with_config_flag_fully_specified_yml(
    monkeypatch, empty_data_context, tmp_path_factory
):
    """We don't care about the NOUN here just combinations of the config flag"""
    context = empty_data_context
    runner = CliRunner(mix_stderr=True)
    temp_dir = tmp_path_factory.mktemp("config_flag_check")
    monkeypatch.chdir(temp_dir)
    result = runner.invoke(
        cli,
        f"--config {context.root_directory}/great_expectations.yml --v3-api checkpoint list",
        catch_exceptions=False,
    )
    assert result.exit_code == 0
    assert "No Checkpoints found" in result.output


def test_cli_works_from_random_directory_with_config_flag_great_expectations_directory(
    monkeypatch, empty_data_context, tmp_path_factory
):
    """We don't care about the NOUN here just combinations of the config flag"""
    context = empty_data_context
    runner = CliRunner(mix_stderr=True)
    temp_dir = tmp_path_factory.mktemp("config_flag_check")
    monkeypatch.chdir(temp_dir)
    result = runner.invoke(
        cli,
        f"--config {context.root_directory} --v3-api checkpoint list",
        catch_exceptions=False,
    )
    assert result.exit_code == 0
    assert "No Checkpoints found" in result.output


def test_cli_works_from_random_directory_with_c_flag_fully_specified_yml(
    monkeypatch, empty_data_context, tmp_path_factory
):
    """We don't care about the NOUN here just combinations of the config flag"""
    context = empty_data_context
    runner = CliRunner(mix_stderr=True)
    temp_dir = tmp_path_factory.mktemp("config_flag_check")
    monkeypatch.chdir(temp_dir)
    result = runner.invoke(
        cli,
        f"-c {context.root_directory}/great_expectations.yml --v3-api checkpoint list",
        catch_exceptions=False,
    )
    assert result.exit_code == 0
    assert "No Checkpoints found" in result.output


def test_cli_works_from_random_directory_with_c_flag_great_expectations_directory(
    monkeypatch, empty_data_context, tmp_path_factory
):
    """We don't care about the NOUN here just combinations of the config flag"""
    context = empty_data_context
    runner = CliRunner(mix_stderr=True)
    temp_dir = tmp_path_factory.mktemp("config_flag_check")
    monkeypatch.chdir(temp_dir)
    result = runner.invoke(
        cli,
        f"-c {context.root_directory} --v3-api checkpoint list",
        catch_exceptions=False,
    )
    assert result.exit_code == 0
    assert "No Checkpoints found" in result.output


CONFIG_NOT_FOUND_ERROR_MESSAGE = "No great_expectations directory was found here!"


def test_cli_config_not_found_raises_error_for_datasource_list(
    tmp_path_factory, monkeypatch
):
    tmp_dir = str(tmp_path_factory.mktemp("test_cli_config_not_found"))
    monkeypatch.chdir(tmp_dir)
    runner = CliRunner(mix_stderr=True)
    result = runner.invoke(
        cli, ["-c", "./", "--v3-api", "datasource", "list"], catch_exceptions=False
    )
    assert CONFIG_NOT_FOUND_ERROR_MESSAGE in result.output
    result = runner.invoke(
        cli, ["--v3-api", "datasource", "list"], catch_exceptions=False
    )
    assert CONFIG_NOT_FOUND_ERROR_MESSAGE in result.output


def test_cli_config_not_found_raises_error_for_datasource_new(
    tmp_path_factory, monkeypatch
):
    tmp_dir = str(tmp_path_factory.mktemp("test_cli_config_not_found"))
    monkeypatch.chdir(tmp_dir)
    runner = CliRunner(mix_stderr=True)
    result = runner.invoke(
        cli, ["-c", "./", "--v3-api", "datasource", "new"], catch_exceptions=False
    )
    assert CONFIG_NOT_FOUND_ERROR_MESSAGE in result.output
    result = runner.invoke(
        cli, ["--v3-api", "datasource", "new"], catch_exceptions=False
    )
    assert CONFIG_NOT_FOUND_ERROR_MESSAGE in result.output


def test_cli_config_not_found_raises_error_for_datasource_delete(
    tmp_path_factory, monkeypatch
):
    tmp_dir = str(tmp_path_factory.mktemp("test_cli_config_not_found"))
    monkeypatch.chdir(tmp_dir)
    runner = CliRunner(mix_stderr=True)
    result = runner.invoke(
        cli,
        ["-c", "./", "--v3-api", "datasource", "delete", "new"],
        catch_exceptions=False,
    )
    assert CONFIG_NOT_FOUND_ERROR_MESSAGE in result.output
    result = runner.invoke(
        cli, ["--v3-api", "datasource", "delete", "new"], catch_exceptions=False
    )
    assert CONFIG_NOT_FOUND_ERROR_MESSAGE in result.output


def test_cli_config_not_found_raises_error_for_project_check_config(
    tmp_path_factory, monkeypatch
):
    tmp_dir = str(tmp_path_factory.mktemp("test_cli_config_not_found"))
    monkeypatch.chdir(tmp_dir)
    runner = CliRunner(mix_stderr=True)
    result = runner.invoke(
        cli,
        ["-c", "./", "--v3-api", "project", "check-config"],
        catch_exceptions=False,
    )
    assert CONFIG_NOT_FOUND_ERROR_MESSAGE in result.output
    result = runner.invoke(
        cli, ["--v3-api", "project", "check-config"], catch_exceptions=False
    )
    assert CONFIG_NOT_FOUND_ERROR_MESSAGE in result.output


def test_cli_config_not_found_raises_error_for_project_upgrade(
    tmp_path_factory, monkeypatch
):
    tmp_dir = str(tmp_path_factory.mktemp("test_cli_config_not_found"))
    monkeypatch.chdir(tmp_dir)
    runner = CliRunner(mix_stderr=True)
    result = runner.invoke(
        cli,
        ["-c", "./", "--v3-api", "project", "upgrade"],
        catch_exceptions=False,
    )
    assert CONFIG_NOT_FOUND_ERROR_MESSAGE in result.output
    result = runner.invoke(
        cli, ["--v3-api", "project", "upgrade"], catch_exceptions=False
    )
    assert CONFIG_NOT_FOUND_ERROR_MESSAGE in result.output


def test_cli_config_not_found_raises_error_for_store_list(
    tmp_path_factory, monkeypatch
):
    tmp_dir = str(tmp_path_factory.mktemp("test_cli_config_not_found"))
    monkeypatch.chdir(tmp_dir)
    runner = CliRunner(mix_stderr=True)
    result = runner.invoke(
        cli, ["-c", "./", "--v3-api", "store", "list"], catch_exceptions=False
    )
    assert CONFIG_NOT_FOUND_ERROR_MESSAGE in result.output
    result = runner.invoke(cli, ["--v3-api", "store", "list"], catch_exceptions=False)
    assert CONFIG_NOT_FOUND_ERROR_MESSAGE in result.output


def test_cli_config_not_found_raises_error_for_suite_new(tmp_path_factory, monkeypatch):
    tmp_dir = str(tmp_path_factory.mktemp("test_cli_config_not_found"))
    monkeypatch.chdir(tmp_dir)
    runner = CliRunner(mix_stderr=True)
    result = runner.invoke(
        cli, ["-c", "./", "--v3-api", "suite", "new"], catch_exceptions=False
    )
    assert CONFIG_NOT_FOUND_ERROR_MESSAGE in result.output
    result = runner.invoke(cli, ["--v3-api", "suite", "new"], catch_exceptions=False)
    assert CONFIG_NOT_FOUND_ERROR_MESSAGE in result.output


def test_cli_config_not_found_raises_error_for_suite_list(
    tmp_path_factory, monkeypatch
):
    tmp_dir = str(tmp_path_factory.mktemp("test_cli_config_not_found"))
    monkeypatch.chdir(tmp_dir)
    runner = CliRunner(mix_stderr=True)
    result = runner.invoke(
        cli, ["-c", "./", "--v3-api", "suite", "list"], catch_exceptions=False
    )
    assert CONFIG_NOT_FOUND_ERROR_MESSAGE in result.output
    result = runner.invoke(cli, ["--v3-api", "suite", "list"], catch_exceptions=False)
    assert CONFIG_NOT_FOUND_ERROR_MESSAGE in result.output


def test_cli_config_not_found_raises_error_for_suite_scaffold(
    tmp_path_factory, monkeypatch
):
    tmp_dir = str(tmp_path_factory.mktemp("test_cli_config_not_found"))
    monkeypatch.chdir(tmp_dir)
    runner = CliRunner(mix_stderr=True)
    result = runner.invoke(
        cli,
        ["-c", "./", "--v3-api", "suite", "new", "--profile", "--interactive"],
        catch_exceptions=False,
    )
    assert CONFIG_NOT_FOUND_ERROR_MESSAGE in result.output
    result = runner.invoke(
        cli,
        ["--v3-api", "suite", "new", "--profile", "--interactive"],
        catch_exceptions=False,
    )
    assert CONFIG_NOT_FOUND_ERROR_MESSAGE in result.output


def test_cli_config_not_found_raises_error_for_suite_edit(
    tmp_path_factory, monkeypatch
):
    tmp_dir = str(tmp_path_factory.mktemp("test_cli_config_not_found"))
    monkeypatch.chdir(tmp_dir)
    runner = CliRunner(mix_stderr=True)
    result = runner.invoke(
        cli, ["-c", "./", "--v3-api", "suite", "edit", "FAKE"], catch_exceptions=False
    )
    assert CONFIG_NOT_FOUND_ERROR_MESSAGE in result.output
    result = runner.invoke(
        cli, ["--v3-api", "suite", "edit", "FAKE"], catch_exceptions=False
    )
    assert CONFIG_NOT_FOUND_ERROR_MESSAGE in result.output


def test_cli_config_not_found_raises_error_for_suite_delete(
    tmp_path_factory, monkeypatch
):
    tmp_dir = str(tmp_path_factory.mktemp("test_cli_config_not_found"))
    monkeypatch.chdir(tmp_dir)
    runner = CliRunner(mix_stderr=True)
    result = runner.invoke(
        cli, ["-c", "./", "suite", "delete", "deleteme"], catch_exceptions=False
    )
    assert CONFIG_NOT_FOUND_ERROR_MESSAGE in result.output
    result = runner.invoke(
        cli, ["--v3-api", "suite", "delete", "deleteme"], catch_exceptions=False
    )
    assert CONFIG_NOT_FOUND_ERROR_MESSAGE in result.output


def test_cli_config_not_found_raises_error_for_checkpoint_new(
    tmp_path_factory, monkeypatch
):
    tmp_dir = str(tmp_path_factory.mktemp("test_cli_config_not_found"))
    monkeypatch.chdir(tmp_dir)
    runner = CliRunner(mix_stderr=True)
    result = runner.invoke(
        cli, ["-c", "./", "--v3-api", "checkpoint", "new"], catch_exceptions=False
    )
    assert CONFIG_NOT_FOUND_ERROR_MESSAGE in result.output
    result = runner.invoke(
        cli, ["--v3-api", "checkpoint", "new"], catch_exceptions=False
    )
    assert CONFIG_NOT_FOUND_ERROR_MESSAGE in result.output


def test_cli_config_not_found_raises_error_for_checkpoint_list(
    tmp_path_factory, monkeypatch
):
    tmp_dir = str(tmp_path_factory.mktemp("test_cli_config_not_found"))
    monkeypatch.chdir(tmp_dir)
    runner = CliRunner(mix_stderr=True)
    result = runner.invoke(
        cli, ["-c", "./", "--v3-api", "checkpoint", "list"], catch_exceptions=False
    )
    assert CONFIG_NOT_FOUND_ERROR_MESSAGE in result.output
    result = runner.invoke(
        cli, ["--v3-api", "checkpoint", "list"], catch_exceptions=False
    )
    assert CONFIG_NOT_FOUND_ERROR_MESSAGE in result.output


def test_cli_config_not_found_raises_error_for_checkpoint_delete(
    tmp_path_factory, monkeypatch
):
    tmp_dir = str(tmp_path_factory.mktemp("test_cli_config_not_found"))
    monkeypatch.chdir(tmp_dir)
    runner = CliRunner(mix_stderr=True)
    result = runner.invoke(
        cli,
        ["-c", "./", "--v3-api", "checkpoint", "delete", "deleteme"],
        catch_exceptions=False,
    )
    assert CONFIG_NOT_FOUND_ERROR_MESSAGE in result.output
    result = runner.invoke(
        cli, ["--v3-api", "checkpoint", "delete", "deleteme"], catch_exceptions=False
    )
    assert CONFIG_NOT_FOUND_ERROR_MESSAGE in result.output


def test_cli_config_not_found_raises_error_for_docs_clean(
    tmp_path_factory, monkeypatch
):
    tmp_dir = str(tmp_path_factory.mktemp("test_cli_config_not_found"))
    monkeypatch.chdir(tmp_dir)
    runner = CliRunner(mix_stderr=True)
    result = runner.invoke(
        cli,
        [
            "-c",
            "./",
            "--v3-api",
            "docs",
            "clean",
        ],
        catch_exceptions=False,
    )
    assert CONFIG_NOT_FOUND_ERROR_MESSAGE in result.output
    result = runner.invoke(cli, ["--v3-api", "docs", "clean"], catch_exceptions=False)
    assert CONFIG_NOT_FOUND_ERROR_MESSAGE in result.output


def test_cli_config_not_found_raises_error_for_docs_list(tmp_path_factory, monkeypatch):
    tmp_dir = str(tmp_path_factory.mktemp("test_cli_config_not_found"))
    monkeypatch.chdir(tmp_dir)
    runner = CliRunner(mix_stderr=True)
    result = runner.invoke(
        cli,
        [
            "-c",
            "./",
            "--v3-api",
            "docs",
            "list",
        ],
        catch_exceptions=False,
    )
    assert CONFIG_NOT_FOUND_ERROR_MESSAGE in result.output
    result = runner.invoke(cli, ["--v3-api", "docs", "list"], catch_exceptions=False)
    assert CONFIG_NOT_FOUND_ERROR_MESSAGE in result.output


def test_cli_config_not_found_raises_error_for_docs_build(
    tmp_path_factory, monkeypatch
):
    tmp_dir = str(tmp_path_factory.mktemp("test_cli_config_not_found"))
    monkeypatch.chdir(tmp_dir)
    runner = CliRunner(mix_stderr=True)
    result = runner.invoke(
        cli,
        ["-c", "./", "--v3-api", "docs", "build", "--no-view"],
        catch_exceptions=False,
    )
    assert CONFIG_NOT_FOUND_ERROR_MESSAGE in result.output

    result = runner.invoke(
        cli,
        ["-c", "./", "--v3-api", "docs", "build", "--no-view"],
        catch_exceptions=False,
    )
    assert CONFIG_NOT_FOUND_ERROR_MESSAGE in result.output
    result = runner.invoke(
        cli, ["--v3-api", "docs", "build", "--no-view"], catch_exceptions=False
    )
    assert CONFIG_NOT_FOUND_ERROR_MESSAGE in result.output


@mock.patch(
    "great_expectations.core.usage_statistics.usage_statistics.UsageStatisticsHandler.emit"
)
def test_assume_yes_using_full_flag_using_checkpoint_delete(
    mock_emit,
    caplog,
    monkeypatch,
    empty_context_with_checkpoint_v1_stats_enabled,
):
    """
    What does this test and why?
    All versions of the --assume-yes flag (--assume-yes/--yes/-y) should behave the same.
    """
    context: DataContext = empty_context_with_checkpoint_v1_stats_enabled
    monkeypatch.chdir(os.path.dirname(context.root_directory))
    runner: CliRunner = CliRunner(mix_stderr=False)
    checkpoint_name: str = "my_v1_checkpoint"
    result: Result = runner.invoke(
        cli,
        f"--v3-api --assume-yes checkpoint delete {checkpoint_name}",
        catch_exceptions=False,
    )
    stdout: str = result.stdout
    assert result.exit_code == 0

    assert (
        f'Are you sure you want to delete the Checkpoint "{checkpoint_name}" (this action is irreversible)?'
        not in stdout
    )
    # This assertion is extra assurance since this test is too permissive if we change the confirmation message
    assert "[Y/n]" not in stdout

    assert 'Checkpoint "my_v1_checkpoint" deleted.' in stdout

    expected_call_args_list = [
        mock.call(
            {"event_payload": {}, "event": "data_context.__init__", "success": True}
        ),
        mock.call(
            {
                "event": "cli.checkpoint.delete.begin",
                "event_payload": {"api_version": "v3"},
                "success": True,
            }
        ),
        mock.call(
            {
                "event": "cli.checkpoint.delete.end",
                "event_payload": {"api_version": "v3"},
                "success": True,
            }
        ),
    ]

    assert mock_emit.call_count == len(expected_call_args_list)
    assert mock_emit.call_args_list == expected_call_args_list

    assert_no_logging_messages_or_tracebacks(
        caplog,
        result,
    )

    result = runner.invoke(
        cli,
        "--v3-api checkpoint list",
        catch_exceptions=False,
    )
    stdout = result.stdout
    assert result.exit_code == 0
    assert "No Checkpoints found." in stdout


@mock.patch(
    "great_expectations.core.usage_statistics.usage_statistics.UsageStatisticsHandler.emit"
)
def test_assume_yes_using_yes_flag_using_checkpoint_delete(
    mock_emit,
    caplog,
    monkeypatch,
    empty_context_with_checkpoint_v1_stats_enabled,
):
    """
    What does this test and why?
    All versions of the --assume-yes flag (--assume-yes/--yes/-y) should behave the same.
    """
    context: DataContext = empty_context_with_checkpoint_v1_stats_enabled
    monkeypatch.chdir(os.path.dirname(context.root_directory))
    runner: CliRunner = CliRunner(mix_stderr=False)
    checkpoint_name: str = "my_v1_checkpoint"
    result: Result = runner.invoke(
        cli,
        f"--v3-api --yes checkpoint delete {checkpoint_name}",
        catch_exceptions=False,
    )
    stdout: str = result.stdout
    assert result.exit_code == 0

    assert (
        f'Are you sure you want to delete the Checkpoint "{checkpoint_name}" (this action is irreversible)?'
        not in stdout
    )
    # This assertion is extra assurance since this test is too permissive if we change the confirmation message
    assert "[Y/n]" not in stdout

    assert 'Checkpoint "my_v1_checkpoint" deleted.' in stdout

    expected_call_args_list = [
        mock.call(
            {"event_payload": {}, "event": "data_context.__init__", "success": True}
        ),
        mock.call(
            {
                "event": "cli.checkpoint.delete.begin",
                "event_payload": {"api_version": "v3"},
                "success": True,
            }
        ),
        mock.call(
            {
                "event": "cli.checkpoint.delete.end",
                "event_payload": {"api_version": "v3"},
                "success": True,
            }
        ),
    ]

    assert mock_emit.call_count == len(expected_call_args_list)
    assert mock_emit.call_args_list == expected_call_args_list

    assert_no_logging_messages_or_tracebacks(
        caplog,
        result,
    )

    result = runner.invoke(
        cli,
        "--v3-api checkpoint list",
        catch_exceptions=False,
    )
    stdout = result.stdout
    assert result.exit_code == 0
    assert "No Checkpoints found." in stdout


@mock.patch(
    "great_expectations.core.usage_statistics.usage_statistics.UsageStatisticsHandler.emit"
)
def test_assume_yes_using_y_flag_using_checkpoint_delete(
    mock_emit,
    caplog,
    monkeypatch,
    empty_context_with_checkpoint_v1_stats_enabled,
):
    """
    What does this test and why?
    All versions of the --assume-yes flag (--assume-yes/--yes/-y) should behave the same.
    """
    context: DataContext = empty_context_with_checkpoint_v1_stats_enabled
    monkeypatch.chdir(os.path.dirname(context.root_directory))
    runner: CliRunner = CliRunner(mix_stderr=False)
    checkpoint_name: str = "my_v1_checkpoint"
    result: Result = runner.invoke(
        cli,
        f"--v3-api -y checkpoint delete {checkpoint_name}",
        catch_exceptions=False,
    )
    stdout: str = result.stdout
    assert result.exit_code == 0

    assert (
        f'Are you sure you want to delete the Checkpoint "{checkpoint_name}" (this action is irreversible)?'
        not in stdout
    )
    # This assertion is extra assurance since this test is too permissive if we change the confirmation message
    assert "[Y/n]" not in stdout

    assert 'Checkpoint "my_v1_checkpoint" deleted.' in stdout

    expected_call_args_list = [
        mock.call(
            {"event_payload": {}, "event": "data_context.__init__", "success": True}
        ),
        mock.call(
            {
                "event": "cli.checkpoint.delete.begin",
                "event_payload": {"api_version": "v3"},
                "success": True,
            }
        ),
        mock.call(
            {
                "event": "cli.checkpoint.delete.end",
                "event_payload": {"api_version": "v3"},
                "success": True,
            }
        ),
    ]

    assert mock_emit.call_count == len(expected_call_args_list)
    assert mock_emit.call_args_list == expected_call_args_list

    assert_no_logging_messages_or_tracebacks(
        caplog,
        result,
    )

    result = runner.invoke(
        cli,
        "--v3-api checkpoint list",
        catch_exceptions=False,
    )
    stdout = result.stdout
    assert result.exit_code == 0
    assert "No Checkpoints found." in stdout


@mock.patch(
    "great_expectations.core.usage_statistics.usage_statistics.UsageStatisticsHandler.emit"
)
def test_using_assume_yes_flag_on_command_with_no_assume_yes_implementation(
    mock_emit,
    caplog,
    monkeypatch,
    titanic_pandas_data_context_with_v013_datasource_stats_enabled_with_checkpoints_v1_with_templates,
):
    """
    What does this test and why?
    The --assume-yes flag should not cause issues when run with commands that do not implement any logic based on it.
    """
    context: DataContext = titanic_pandas_data_context_with_v013_datasource_stats_enabled_with_checkpoints_v1_with_templates
    monkeypatch.chdir(os.path.dirname(context.root_directory))
    runner: CliRunner = CliRunner(mix_stderr=False)
    result: Result = runner.invoke(
        cli,
        "--v3-api --assume-yes checkpoint list",
        catch_exceptions=False,
    )
    stdout: str = result.stdout
    assert result.exit_code == 0
    assert "Found 8 Checkpoints." in stdout
    checkpoint_names_list: List[str] = [
        "my_simple_checkpoint_with_slack_and_notify_with_all",
        "my_nested_checkpoint_template_1",
        "my_nested_checkpoint_template_3",
        "my_nested_checkpoint_template_2",
        "my_simple_checkpoint_with_site_names",
        "my_minimal_simple_checkpoint",
        "my_simple_checkpoint_with_slack",
        "my_simple_template_checkpoint",
    ]
    assert all([checkpoint_name in stdout for checkpoint_name in checkpoint_names_list])

    expected_call_args_list = [
        mock.call(
            {"event_payload": {}, "event": "data_context.__init__", "success": True}
        ),
        mock.call(
            {
                "event": "cli.checkpoint.list.begin",
                "event_payload": {"api_version": "v3"},
                "success": True,
            }
        ),
        mock.call(
            {
                "event": "cli.checkpoint.list.end",
                "event_payload": {"api_version": "v3"},
                "success": True,
            }
        ),
    ]

    assert mock_emit.call_count == len(expected_call_args_list)
    assert mock_emit.call_args_list == expected_call_args_list

    assert_no_logging_messages_or_tracebacks(
        caplog,
        result,
    )



from __future__ import annotations

from enum import Enum
from typing import Any, List, Union


class GXSqlDialect(Enum):
    """Contains sql dialects that have some level of support in Great Expectations.
    Also contains an unsupported attribute if the dialect is not in the list.
    """

    AWSATHENA = "awsathena"
    BIGQUERY = "bigquery"
    DREMIO = "dremio"
    HIVE = "hive"
    MSSQL = "mssql"
    MYSQL = "mysql"
    ORACLE = "oracle"
    POSTGRESQL = "postgresql"
    REDSHIFT = "redshift"
    SNOWFLAKE = "snowflake"
    SQLITE = "sqlite"
    TERADATASQL = "teradatasql"
    TRINO = "trino"
    VERTICA = "vertica"
    OTHER = "other"

    def __eq__(self, other: Union[str, bytes, GXSqlDialect]):  # type: ignore[override] # supertype uses `object`
        if isinstance(other, str):
            return self.value.lower() == other.lower()
        # Comparison against byte string, e.g. `b"hive"` should be treated as unicode
        elif isinstance(other, bytes):
            return self.value.lower() == other.lower().decode("utf-8")
        return self.value.lower() == other.value.lower()

    def __hash__(self: GXSqlDialect):
        return hash(self.value)

    @classmethod
    def _missing_(cls, value: Any) -> Any:
        try:
            # Sometimes `value` is a byte string, e.g. `b"hive"`, it should be converted
            return cls(value.decode())
        except (UnicodeDecodeError, AttributeError):
            return super()._missing_(value)

    @classmethod
    def get_all_dialect_names(cls) -> List[str]:
        """Get dialect names for all SQL dialects."""
        return [
            dialect_name.value
            for dialect_name in cls
            if dialect_name != GXSqlDialect.OTHER
        ]

    @classmethod
    def get_all_dialects(cls) -> List[GXSqlDialect]:
        """Get all dialects."""
        return [dialect for dialect in cls if dialect != GXSqlDialect.OTHER]



from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Dict, List, Optional, Set, Type

from great_expectations.rule_based_profiler.data_assistant import (
    DataAssistant,  # noqa: TCH001
)
from great_expectations.rule_based_profiler.data_assistant.data_assistant_runner import (
    DataAssistantRunner,
)

if TYPE_CHECKING:
    from great_expectations.data_context.data_context.abstract_data_context import (
        AbstractDataContext,
    )

logger = logging.getLogger(__name__)


class DataAssistantDispatcher:
    """
    DataAssistantDispatcher intercepts requests for "DataAssistant" classes by their registered names and manages their
    associated "DataAssistantRunner" objects, which process invocations of calls to "DataAssistant" "run()" methods.
    """

    _registered_data_assistants: Dict[str, Type[DataAssistant]] = {}

    def __init__(self, data_context: AbstractDataContext) -> None:
        """
        Args:
            data_context: AbstractDataContext associated with DataAssistantDispatcher
        """
        self._data_context = data_context

        self._data_assistant_runner_cache: dict = {}

    def __getattr__(self, name: str) -> DataAssistantRunner:
        # Both, registered data_assistant_type and alias name are supported for invocation.

        # _registered_data_assistants has both aliases and full names
        data_assistant_cls: Optional[
            Type[DataAssistant]
        ] = DataAssistantDispatcher._get_data_assistant_impl(name=name)

        # If "DataAssistant" is not registered, then raise "AttributeError", which is appropriate for "__getattr__()".
        if data_assistant_cls is None:
            raise AttributeError(
                f'"{type(self).__name__}" object has no attribute "{name}".'
            )

        data_assistant_name: str = data_assistant_cls.data_assistant_type
        data_assistant_runner: Optional[
            DataAssistantRunner
        ] = self._data_assistant_runner_cache.get(data_assistant_name)
        if data_assistant_runner is None:
            data_assistant_runner = DataAssistantRunner(
                data_assistant_cls=data_assistant_cls,
                data_context=self._data_context,
            )
            self._data_assistant_runner_cache[
                data_assistant_name
            ] = data_assistant_runner

        return data_assistant_runner

    @classmethod
    def _register_data_assistant(
        cls,
        data_assistant: Type[DataAssistant],
    ) -> None:
        """
        This method registers "DataAssistant" subclass for future instantiation and execution of its "run()" method.

        Args:
            data_assistant: "DataAssistant" class to be registered
        """
        # TODO: <Alex>6/21/2022: Only alias is to be registered (leave formal type registration in as commented).</Alex>
        # data_assistant_type = data_assistant.data_assistant_type
        # cls._register(data_assistant_type, data_assistant)

        alias: Optional[str] = data_assistant.__alias__
        if alias is not None:
            cls._register(alias, data_assistant)

    @classmethod
    def _register(cls, name: str, data_assistant: Type[DataAssistant]) -> None:
        registered_data_assistants = cls._registered_data_assistants

        if name in registered_data_assistants:
            raise ValueError(f'Existing declarations of DataAssistant "{name}" found.')

        logger.debug(
            f'Registering the declaration of DataAssistant "{name}" took place.'
        )
        registered_data_assistants[name] = data_assistant

    @classmethod
    def _get_data_assistant_impl(
        cls,
        name: Optional[str],
    ) -> Optional[Type[DataAssistant]]:
        """
        This method obtains (previously registered) "DataAssistant" class from DataAssistant Registry.

        Note that it will clean the input string before checking against registered assistants.

        Args:
            name: String representing "snake case" version of "DataAssistant" class type

        Returns:
            Class inheriting "DataAssistant" if found; otherwise, None
        """
        if name is None:
            return None

        name = name.lower()

        return cls._registered_data_assistants.get(name)

    def __dir__(self) -> List[str]:
        """
        This custom magic method is used to enable tab completion on "DataAssistantDispatcher" objects.
        """
        data_assistant_dispatcher_attrs: Set[str] = set(super().__dir__())
        data_assistant_registered_names: Set[
            str
        ] = get_registered_data_assistant_names()
        combined_dir_attrs: Set[str] = (
            data_assistant_dispatcher_attrs | data_assistant_registered_names
        )
        return list(combined_dir_attrs)


def get_registered_data_assistant_names() -> Set[str]:
    """
    This method returns names (registered data_assistant_type and alias name) of registered "DataAssistant" classes.
    """
    return set(DataAssistantDispatcher._registered_data_assistants.keys())



from typing import List

from ruamel import yaml

from great_expectations import DataContext
from great_expectations.core import ExpectationConfiguration
from great_expectations.rule_based_profiler import RuleBasedProfilerResult
from great_expectations.rule_based_profiler.rule_based_profiler import RuleBasedProfiler

profiler_config = r"""
# <snippet name="tests/integration/docusaurus/expectations/advanced/multi_batch_rule_based_profiler_example.py full profiler_config">
# <snippet name="tests/integration/docusaurus/expectations/advanced/multi_batch_rule_based_profiler_example.py full row_count_rule">
# This profiler is meant to be used on the NYC taxi data (yellow_tripdata_sample_<YEAR>-<MONTH>.csv)
# located in tests/test_sets/taxi_yellow_tripdata_samples/

# <snippet name="tests/integration/docusaurus/expectations/advanced/multi_batch_rule_based_profiler_example.py name and config_version">
name: My Profiler
config_version: 1.0
# </snippet>

# <snippet name="tests/integration/docusaurus/expectations/advanced/multi_batch_rule_based_profiler_example.py variables and rule name">
variables:
  false_positive_rate: 0.01
  mostly: 1.0

rules:
  row_count_rule:
# </snippet>
# <snippet name="tests/integration/docusaurus/expectations/advanced/multi_batch_rule_based_profiler_example.py row_count_rule domain_builder">
    domain_builder:
        class_name: TableDomainBuilder
# </snippet>
# <snippet name="tests/integration/docusaurus/expectations/advanced/multi_batch_rule_based_profiler_example.py row_count_rule parameter_builders">
    parameter_builders:
      - name: row_count_range
        class_name: NumericMetricRangeMultiBatchParameterBuilder
        metric_name: table.row_count
        metric_domain_kwargs: $domain.domain_kwargs
        false_positive_rate: $variables.false_positive_rate
        truncate_values:
          lower_bound: 0
        round_decimals: 0
# </snippet>
# <snippet name="tests/integration/docusaurus/expectations/advanced/multi_batch_rule_based_profiler_example.py row_count_rule expectation_configuration_builders">
    expectation_configuration_builders:
      - expectation_type: expect_table_row_count_to_be_between
        class_name: DefaultExpectationConfigurationBuilder
        module_name: great_expectations.rule_based_profiler.expectation_configuration_builder
        min_value: $parameter.row_count_range.value[0]
        max_value: $parameter.row_count_range.value[1]
        mostly: $variables.mostly
        meta:
          profiler_details: $parameter.row_count_range.details
# </snippet>
# </snippet>
# <snippet name="tests/integration/docusaurus/expectations/advanced/multi_batch_rule_based_profiler_example.py column_ranges_rule domain_builder">
  column_ranges_rule:
    domain_builder:
      class_name: ColumnDomainBuilder
      include_semantic_types:
        - numeric
# </snippet>
# <snippet name="tests/integration/docusaurus/expectations/advanced/multi_batch_rule_based_profiler_example.py column_ranges_rule parameter_builders">
    parameter_builders:
      - name: min_range
        class_name: NumericMetricRangeMultiBatchParameterBuilder
        metric_name: column.min
        metric_domain_kwargs: $domain.domain_kwargs
        false_positive_rate: $variables.false_positive_rate
        round_decimals: 2
      - name: max_range
        class_name: NumericMetricRangeMultiBatchParameterBuilder
        metric_name: column.max
        metric_domain_kwargs: $domain.domain_kwargs
        false_positive_rate: $variables.false_positive_rate
        round_decimals: 2
# </snippet>
# <snippet name="tests/integration/docusaurus/expectations/advanced/multi_batch_rule_based_profiler_example.py column_ranges_rule expectation_configuration_builders">
    expectation_configuration_builders:
      - expectation_type: expect_column_min_to_be_between
        class_name: DefaultExpectationConfigurationBuilder
        module_name: great_expectations.rule_based_profiler.expectation_configuration_builder
        column: $domain.domain_kwargs.column
        min_value: $parameter.min_range.value[0]
        max_value: $parameter.min_range.value[1]
        mostly: $variables.mostly
        meta:
          profiler_details: $parameter.min_range.details
      - expectation_type: expect_column_max_to_be_between
        class_name: DefaultExpectationConfigurationBuilder
        module_name: great_expectations.rule_based_profiler.expectation_configuration_builder
        column: $domain.domain_kwargs.column
        min_value: $parameter.max_range.value[0]
        max_value: $parameter.max_range.value[1]
        mostly: $variables.mostly
        meta:
          profiler_details: $parameter.max_range.details
# </snippet>
# </snippet>
"""


# <snippet name="tests/integration/docusaurus/expectations/advanced/multi_batch_rule_based_profiler_example.py set up">
data_context = DataContext()

full_profiler_config_dict: dict = yaml.load(profiler_config)
# </snippet>

# Instantiate RuleBasedProfiler
# <snippet name="tests/integration/docusaurus/expectations/advanced/multi_batch_rule_based_profiler_example.py instantiate">
rule_based_profiler: RuleBasedProfiler = RuleBasedProfiler(
    name=full_profiler_config_dict["name"],
    config_version=full_profiler_config_dict["config_version"],
    rules=full_profiler_config_dict["rules"],
    variables=full_profiler_config_dict["variables"],
    data_context=data_context,
)
# </snippet>


# <snippet name="tests/integration/docusaurus/expectations/advanced/multi_batch_rule_based_profiler_example.py run">
batch_request: dict = {
    "datasource_name": "taxi_pandas",
    "data_connector_name": "monthly",
    "data_asset_name": "my_reports",
    "data_connector_query": {
        "index": "-6:-1",
    },
}

result: RuleBasedProfilerResult = rule_based_profiler.run(batch_request=batch_request)
# </snippet>

expectation_configurations: List[
    ExpectationConfiguration
] = result.expectation_configurations

print(expectation_configurations)

# Please note that this docstring is here to demonstrate output for docs. It is not needed for normal use.
# <snippet name="tests/integration/docusaurus/expectations/advanced/multi_batch_rule_based_profiler_example.py row_count_rule_suite">
row_count_rule_suite = """
    {
        "meta": {"great_expectations_version": "0.13.19+58.gf8a650720.dirty"},
        "data_asset_type": None,
        "expectations": [
            {
                "kwargs": {"min_value": 10000, "max_value": 10000, "mostly": 1.0},
                "expectation_type": "expect_table_row_count_to_be_between",
                "meta": {
                    "profiler_details": {
                        "metric_configuration": {
                            "metric_name": "table.row_count",
                            "metric_domain_kwargs": {},
                        }
                    }
                },
            }
        ],
        "expectation_suite_name": "tmp_suite_Profiler_e66f7cbb",
    }
"""
# </snippet>



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Iterator

import pandas as pd

from great_expectations.core.metric_domain_types import MetricDomainTypes
from great_expectations.execution_engine import (
    PandasExecutionEngine,
    SparkDFExecutionEngine,
    SqlAlchemyExecutionEngine,
)
from great_expectations.execution_engine.sqlalchemy_dialect import GXSqlDialect
from great_expectations.expectations.metrics.import_manager import sa
from great_expectations.expectations.metrics.metric_provider import metric_value
from great_expectations.expectations.metrics.table_metric_provider import (
    TableMetricProvider,
)
from great_expectations.validator.metric_configuration import MetricConfiguration
from great_expectations.validator.validator import Validator

if TYPE_CHECKING:
    from great_expectations.expectations.metrics.import_manager import pyspark_sql_Row


class TableHead(TableMetricProvider):
    metric_name = "table.head"
    value_keys = ("n_rows", "fetch_all")
    default_kwarg_values = {"n_rows": 5, "fetch_all": False}

    @metric_value(engine=PandasExecutionEngine)
    def _pandas(
        cls,
        execution_engine: PandasExecutionEngine,
        metric_domain_kwargs: dict,
        metric_value_kwargs: dict,
        metrics: dict[str, Any],
        runtime_configuration: dict,
    ) -> pd.DataFrame:
        df, _, _ = execution_engine.get_compute_domain(
            metric_domain_kwargs, domain_type=MetricDomainTypes.TABLE
        )
        if metric_value_kwargs.get("fetch_all", cls.default_kwarg_values["fetch_all"]):
            return df
        n_rows: int = (
            metric_value_kwargs.get("n_rows")
            if metric_value_kwargs.get("n_rows") is not None
            else cls.default_kwarg_values["n_rows"]
        )
        return df.head(n=n_rows)

    @metric_value(engine=SqlAlchemyExecutionEngine)
    def _sqlalchemy(  # noqa: C901 - 16
        cls,
        execution_engine: SqlAlchemyExecutionEngine,
        metric_domain_kwargs: dict,
        metric_value_kwargs: dict,
        metrics: dict[str, Any],
        runtime_configuration: dict,
    ) -> pd.DataFrame:
        selectable, _, _ = execution_engine.get_compute_domain(
            metric_domain_kwargs, domain_type=MetricDomainTypes.TABLE
        )
        table_name = getattr(selectable, "name", None)
        n_rows: int = (
            metric_value_kwargs.get("n_rows")
            if metric_value_kwargs.get("n_rows") is not None
            else cls.default_kwarg_values["n_rows"]
        )
        df_chunk_iterator: Iterator[pd.DataFrame]
        if (
            isinstance(table_name, sa.sql.elements._anonymous_label)
            or table_name is None
        ):
            # if a custom query was passed
            try:
                if metric_value_kwargs["fetch_all"]:
                    df = pd.read_sql_query(
                        sql=selectable,
                        con=execution_engine.engine,
                    )
                else:
                    # passing chunksize causes the Iterator to be returned
                    df_chunk_iterator = pd.read_sql_query(
                        sql=selectable,
                        con=execution_engine.engine,
                        chunksize=abs(n_rows),
                    )
                    df = TableHead._get_head_df_from_df_iterator(
                        df_chunk_iterator=df_chunk_iterator, n_rows=n_rows
                    )
            except (ValueError, NotImplementedError):
                # MetaData that is used by pd.read_sql_table
                # cannot work on a temp table with pandas < 1.4.0.
                # If it fails, we try to get the data using read_sql.
                df = None
            except StopIteration:
                validator = Validator(execution_engine=execution_engine)
                columns = validator.get_metric(
                    MetricConfiguration("table.columns", metric_domain_kwargs)
                )
                df = pd.DataFrame(columns=columns)
        else:
            try:
                if metric_value_kwargs["fetch_all"]:
                    df = pd.read_sql_table(
                        table_name=getattr(selectable, "name", None),
                        schema=getattr(selectable, "schema", None),
                        con=execution_engine.engine,
                    )
                else:
                    # passing chunksize causes the Iterator to be returned
                    df_chunk_iterator = pd.read_sql_table(
                        table_name=getattr(selectable, "name", None),
                        schema=getattr(selectable, "schema", None),
                        con=execution_engine.engine,
                        chunksize=abs(n_rows),
                    )
                    df = TableHead._get_head_df_from_df_iterator(
                        df_chunk_iterator=df_chunk_iterator, n_rows=n_rows
                    )

            except (ValueError, NotImplementedError):
                # MetaData that is used by pd.read_sql_table
                # cannot work on a temp table with pandas < 1.4.0.
                # If it fails, we try to get the data using read_sql.
                df = None
            except StopIteration:
                validator = Validator(execution_engine=execution_engine)
                columns = validator.get_metric(
                    MetricConfiguration("table.columns", metric_domain_kwargs)
                )
                df = pd.DataFrame(columns=columns)

        if df is None:
            # we want to compile our selectable
            stmt = sa.select(["*"]).select_from(selectable)
            fetch_all = metric_value_kwargs["fetch_all"]
            if fetch_all:
                sql = stmt.compile(
                    dialect=execution_engine.engine.dialect,
                    compile_kwargs={"literal_binds": True},
                )
            elif execution_engine.engine.dialect.name.lower() == GXSqlDialect.MSSQL:
                # limit doesn't compile properly for mssql
                sql = str(
                    stmt.compile(
                        dialect=execution_engine.engine.dialect,
                        compile_kwargs={"literal_binds": True},
                    )
                )
                if n_rows > 0:
                    sql = f"SELECT TOP {n_rows}{sql[6:]}"
            else:
                if n_rows > 0:
                    stmt = stmt.limit(n_rows)
                sql = stmt.compile(
                    dialect=execution_engine.engine.dialect,
                    compile_kwargs={"literal_binds": True},
                )

            # if read_sql_query or read_sql_table failed, we try to use the read_sql convenience method
            if n_rows <= 0 and not fetch_all:
                df_chunk_iterator = pd.read_sql(
                    sql=sql, con=execution_engine.engine, chunksize=abs(n_rows)
                )
                df = TableHead._get_head_df_from_df_iterator(
                    df_chunk_iterator=df_chunk_iterator, n_rows=n_rows
                )
            else:
                df = pd.read_sql(sql=sql, con=execution_engine.engine)

        return df

    @staticmethod
    def _get_head_df_from_df_iterator(
        df_chunk_iterator: Iterator[pd.DataFrame], n_rows: int
    ) -> pd.DataFrame:
        if n_rows > 0:
            df = next(df_chunk_iterator)
        else:
            # if n_rows is zero or negative, remove the last chunk
            df_chunk_list: list[pd.DataFrame]
            df_last_chunk: pd.DataFrame
            *df_chunk_list, df_last_chunk = df_chunk_iterator
            if df_chunk_list:
                df = pd.concat(objs=df_chunk_list, ignore_index=True)
            else:
                # if n_rows is zero, the last chunk is the entire dataframe,
                # so we truncate it to preserve the header
                df = df_last_chunk.head(0)

        return df

    @metric_value(engine=SparkDFExecutionEngine)
    def _spark(
        cls,
        execution_engine: SparkDFExecutionEngine,
        metric_domain_kwargs: dict,
        metric_value_kwargs: dict,
        metrics: dict[str, Any],
        runtime_configuration: dict,
    ) -> pd.DataFrame:
        df, _, _ = execution_engine.get_compute_domain(
            metric_domain_kwargs, domain_type=MetricDomainTypes.TABLE
        )
        rows: list[pyspark_sql_Row] | pyspark_sql_Row | list[dict]
        if metric_value_kwargs["fetch_all"]:
            rows = df.collect()
        else:
            n_rows: int = (
                metric_value_kwargs.get("n_rows")
                if metric_value_kwargs.get("n_rows") is not None
                else cls.default_kwarg_values["n_rows"]
            )
            if n_rows >= 0:
                rows = df.head(n=n_rows)
            else:
                rows = df.head(n=df.count() + n_rows)

        rows = [element.asDict() for element in rows]
        df = pd.DataFrame(data=rows)

        return df



import os

from ruamel import yaml

import great_expectations as gx

CONNECTION_STRING = "postgresql+psycopg2://postgres:@localhost/test_ci"

# This utility is not for general use. It is only to support testing.
from tests.test_utils import load_data_into_test_database

load_data_into_test_database(
    table_name="titanic",
    csv_path="./data/Titanic.csv",
    connection_string=CONNECTION_STRING,
    load_full_dataset=True,
)

context = gx.get_context()

# parse great_expectations.yml for comparison
great_expectations_yaml_file_path = os.path.join(
    context.root_directory, "great_expectations.yml"
)
with open(great_expectations_yaml_file_path) as f:
    great_expectations_yaml = yaml.safe_load(f)

actual_datasource = great_expectations_yaml["datasources"]

# expected Datasource
expected_existing_datasource_yaml = r"""
  my_postgres_datasource:
    module_name: great_expectations.datasource
    class_name: Datasource
    execution_engine:
      module_name: great_expectations.execution_engine
      class_name: SqlAlchemyExecutionEngine
      connection_string: postgresql+psycopg2://postgres:@localhost/test_ci
    data_connectors:
      default_runtime_data_connector_name:
        module_name: great_expectations.datasource.data_connector
        class_name: RuntimeDataConnector
        batch_identifiers:
          - default_identifier_name
      default_inferred_data_connector_name:
        module_name: great_expectations.datasource.data_connector
        class_name: InferredAssetSqlDataConnector
        include_schema_name: true
"""

assert actual_datasource == yaml.safe_load(expected_existing_datasource_yaml)

# check that checkpoint contains the right configuration
# parse great_expectations.yml for comparison
checkpoint_yaml_file_path = os.path.join(
    context.root_directory, "checkpoints/test_v3_checkpoint.yml"
)
with open(checkpoint_yaml_file_path) as f:
    actual_checkpoint_yaml = yaml.safe_load(f)

expected_checkpoint_yaml = """
name: test_v3_checkpoint
config_version: 1.0 # Note this is the version of the Checkpoint configuration, and not the great_expectations.yml configuration
template_name:
module_name: great_expectations.checkpoint
class_name: Checkpoint
run_name_template: '%Y%m%d-%H%M%S-my-run-name-template'
expectation_suite_name:
batch_request:
action_list:
  - name: store_validation_result
    action:
      class_name: StoreValidationResultAction
  - name: store_evaluation_params
    action:
      class_name: StoreEvaluationParametersAction
  - name: update_data_docs
    action:
      class_name: UpdateDataDocsAction
      site_names: []
evaluation_parameters: {}
runtime_configuration: {}
validations:
  - batch_request:
      datasource_name: my_postgres_datasource
      data_connector_name: default_runtime_data_connector_name
      data_asset_name: titanic
      runtime_parameters:
        query: SELECT * from public.titanic
      batch_identifiers:
        default_identifier_name: default_identifier
    expectation_suite_name: Titanic.profiled
profilers: []
ge_cloud_id:
expectation_suite_ge_cloud_id:
"""

assert actual_checkpoint_yaml == yaml.safe_load(expected_checkpoint_yaml)

# run checkpoint
results = context.run_checkpoint(checkpoint_name="test_v3_checkpoint")
assert results["success"] is True



from typing import Dict

import pytest
from marshmallow.exceptions import ValidationError
from ruamel.yaml.comments import CommentedMap

from great_expectations.rule_based_profiler.config import (
    DomainBuilderConfig,
    DomainBuilderConfigSchema,
    ExpectationConfigurationBuilderConfig,
    ExpectationConfigurationBuilderConfigSchema,
    NotNullSchema,
    ParameterBuilderConfig,
    ParameterBuilderConfigSchema,
    RuleBasedProfilerConfig,
    RuleBasedProfilerConfigSchema,
    RuleConfig,
    RuleConfigSchema,
)
from great_expectations.rule_based_profiler.config.base import (
    ruleBasedProfilerConfigSchema,
)
from great_expectations.rule_based_profiler.rule_based_profiler import RuleBasedProfiler


@pytest.mark.unit
def test_not_null_schema_raises_error_with_improperly_implemented_subclass():
    class MySchema(NotNullSchema):
        pass

    with pytest.raises(NotImplementedError) as e:
        MySchema().load({})

    assert "must define its own custom __config_class__" in str(e.value)


@pytest.mark.unit
def test_not_null_schema_removes_null_values_when_dumping():
    schema = DomainBuilderConfigSchema()
    config = DomainBuilderConfig(
        class_name="DomainBuilder",
        module_name="great_expectations.rule_based_profiler.domain_builder",
        batch_request=None,
    )

    data = schema.dump(config)
    assert isinstance(data, dict)
    assert "class_name" in data and "module_name" in data
    assert "batch_request" not in data


@pytest.mark.unit
def test_domain_builder_config_successfully_loads_with_required_args():
    data = {
        "class_name": "DomainBuilder",
    }
    schema = DomainBuilderConfigSchema()
    config = schema.load(data)
    assert isinstance(config, DomainBuilderConfig)
    assert all(getattr(config, k) == v for k, v in data.items())
    assert config.module_name == "great_expectations.rule_based_profiler.domain_builder"


@pytest.mark.unit
def test_domain_builder_config_successfully_loads_with_optional_args():
    data = {
        "class_name": "DomainBuilder",
        "module_name": "great_expectations.rule_based_profiler.domain_builder",
        "batch_request": {"datasource_name": "my_datasource"},
    }
    schema = DomainBuilderConfigSchema()
    config = schema.load(data)
    assert isinstance(config, DomainBuilderConfig)
    assert all(getattr(config, k) == v for k, v in data.items())


@pytest.mark.unit
def test_parameter_builder_config_successfully_loads_with_required_args():
    data = {"class_name": "ParameterBuilder", "name": "my_parameter_builder"}
    schema = ParameterBuilderConfigSchema()
    config = schema.load(data)
    assert isinstance(config, ParameterBuilderConfig)
    assert all(getattr(config, k) == v for k, v in data.items())
    assert (
        config.module_name == "great_expectations.rule_based_profiler.parameter_builder"
    )


@pytest.mark.unit
def test_parameter_builder_config_successfully_loads_with_optional_args():
    data = {
        "name": "my_parameter_builder",
        "class_name": "ParameterBuilder",
        "module_name": "great_expectations.rule_based_profiler.parameter_builder",
        "batch_request": {"datasource_name": "my_datasource"},
    }
    schema = ParameterBuilderConfigSchema()
    config = schema.load(data)
    assert isinstance(config, ParameterBuilderConfig)
    assert all(getattr(config, k) == v for k, v in data.items())


@pytest.mark.unit
def test_parameter_builder_config_unsuccessfully_loads_with_missing_required_fields():
    data = {}
    schema = ParameterBuilderConfigSchema()

    with pytest.raises(ValidationError) as e:
        schema.load(data)

    assert "'name': ['Missing data for required field.']" in str(e.value)


@pytest.mark.unit
def test_expectation_configuration_builder_config_successfully_loads_with_required_args():
    data = {
        "class_name": "ExpectationConfigurationBuilder",
        "expectation_type": "expect_column_pair_values_A_to_be_greater_than_B",
    }
    schema = ExpectationConfigurationBuilderConfigSchema()
    config = schema.load(data)
    assert isinstance(config, ExpectationConfigurationBuilderConfig)
    assert all(getattr(config, k) == v for k, v in data.items())
    assert (
        config.module_name
        == "great_expectations.rule_based_profiler.expectation_configuration_builder"
    )


@pytest.mark.unit
def test_expectation_configuration_builder_config_successfully_loads_with_optional_args():
    data = {
        "expectation_type": "expect_column_pair_values_A_to_be_greater_than_B",
        "class_name": "ExpectationConfigurationBuilder",
        "module_name": "great_expectations.rule_based_profiler.expectation_configuration_builder",
        "mostly": 0.9,
        "meta": {"foo": "bar"},
    }
    schema = ExpectationConfigurationBuilderConfigSchema()
    config = schema.load(data)
    assert isinstance(config, ExpectationConfigurationBuilderConfig)
    assert all(getattr(config, k) == v for k, v in data.items())


@pytest.mark.unit
def test_expectation_configuration_builder_config_unsuccessfully_loads_with_missing_required_fields():
    data = {}
    schema = ExpectationConfigurationBuilderConfigSchema()

    with pytest.raises(ValidationError) as e:
        schema.load(data)

    assert (
        "'expectation_type': ['expectation_type missing in expectation configuration builder']"
        in str(e.value)
    )


@pytest.mark.unit
def test_rule_config_successfully_loads_with_required_args():
    data = {
        "domain_builder": {"class_name": "DomainBuilder"},
        "parameter_builders": [
            {"class_name": "ParameterBuilder", "name": "my_parameter"}
        ],
        "expectation_configuration_builders": [
            {
                "class_name": "ExpectationConfigurationBuilder",
                "expectation_type": "expect_column_pair_values_A_to_be_greater_than_B",
            }
        ],
    }
    schema = RuleConfigSchema()
    config = schema.load(data)

    assert isinstance(config, RuleConfig)
    assert isinstance(config.domain_builder, DomainBuilderConfig)
    assert len(config.parameter_builders) == 1 and isinstance(
        config.parameter_builders[0], ParameterBuilderConfig
    )
    assert len(config.expectation_configuration_builders) == 1 and isinstance(
        config.expectation_configuration_builders[0],
        ExpectationConfigurationBuilderConfig,
    )


@pytest.mark.unit
def test_rule_config_unsuccessfully_loads_with_missing_required_fields():
    data = {}
    schema = RuleConfigSchema()

    config = schema.load(data)

    assert isinstance(config, RuleConfig)
    assert config.domain_builder is None
    assert config.parameter_builders is None
    assert config.expectation_configuration_builders is None


@pytest.mark.unit
def test_rule_based_profiler_config_successfully_loads_with_required_args():
    data = {
        "name": "my_RBP",
        "config_version": 1.0,
        "rules": {
            "rule_1": {
                "domain_builder": {"class_name": "DomainBuilder"},
                "parameter_builders": [
                    {"class_name": "ParameterBuilder", "name": "my_parameter"}
                ],
                "expectation_configuration_builders": [
                    {
                        "class_name": "ExpectationConfigurationBuilder",
                        "expectation_type": "expect_column_pair_values_A_to_be_greater_than_B",
                    }
                ],
            },
        },
    }
    schema = RuleBasedProfilerConfigSchema()
    config = schema.load(data)
    assert isinstance(config, dict)
    assert len(config["rules"]) == 1 and isinstance(
        config["rules"]["rule_1"], RuleConfig
    )


@pytest.mark.unit
def test_rule_based_profiler_config_successfully_loads_with_optional_args():
    data = {
        "name": "my_RBP",
        "config_version": 1.0,
        "variables": {"foo": "bar"},
        "rules": {
            "rule_1": {
                "domain_builder": {"class_name": "DomainBuilder"},
                "parameter_builders": [
                    {"class_name": "ParameterBuilder", "name": "my_parameter"}
                ],
                "expectation_configuration_builders": [
                    {
                        "class_name": "ExpectationConfigurationBuilder",
                        "expectation_type": "expect_column_pair_values_A_to_be_greater_than_B",
                    }
                ],
            },
        },
    }
    schema = RuleBasedProfilerConfigSchema()
    config = schema.load(data)
    assert isinstance(config, dict)
    assert data["variables"] == config["variables"]


@pytest.mark.unit
def test_rule_based_profiler_config_unsuccessfully_loads_with_missing_required_fields():
    data = {}
    schema = RuleBasedProfilerConfigSchema()

    with pytest.raises(ValidationError) as e:
        schema.load(data)

    assert all(
        f"'{attr}': ['Missing data for required field.']" in str(e.value)
        for attr in (
            "name",
            "config_version",
            "rules",
        )
    )


@pytest.mark.unit
def test_rule_based_profiler_from_commented_map():
    data = {
        "name": "my_RBP",
        "config_version": 1.0,
        "variables": {"foo": "bar"},
        "rules": {
            "rule_1": {
                "domain_builder": {"class_name": "DomainBuilder"},
                "parameter_builders": [
                    {"class_name": "ParameterBuilder", "name": "my_parameter"}
                ],
                "expectation_configuration_builders": [
                    {
                        "class_name": "ExpectationConfigurationBuilder",
                        "expectation_type": "expect_column_pair_values_A_to_be_greater_than_B",
                    }
                ],
            },
        },
    }
    commented_map = CommentedMap(data)
    config = RuleBasedProfilerConfig.from_commented_map(commented_map)
    assert all(hasattr(config, k) for k in data)


@pytest.mark.unit
def test_resolve_config_using_acceptable_arguments(
    profiler_with_placeholder_args: RuleBasedProfiler,
) -> None:
    old_config: RuleBasedProfilerConfig = profiler_with_placeholder_args.config

    # Roundtrip through schema validation to add/or restore any missing fields.
    old_deserialized_config: dict = ruleBasedProfilerConfigSchema.load(
        old_config.to_json_dict()
    )
    old_deserialized_config.pop("class_name")
    old_deserialized_config.pop("module_name")

    old_config = RuleBasedProfilerConfig(**old_deserialized_config)

    # Brand new config is created but existing attributes are unchanged
    new_config: RuleBasedProfilerConfig = (
        RuleBasedProfilerConfig.resolve_config_using_acceptable_arguments(
            profiler=profiler_with_placeholder_args,
        )
    )

    # Roundtrip through schema validation to add/or restore any missing fields.
    # new_deserialized_config: dict = ruleBasedProfilerConfigSchema.load(new_config.to_json_dict())
    new_deserialized_config: dict = new_config.to_json_dict()
    new_deserialized_config.pop("class_name")
    new_deserialized_config.pop("module_name")

    new_config = RuleBasedProfilerConfig(**new_deserialized_config)

    assert id(old_config) != id(new_config)
    assert all(
        old_config[attr] == new_config[attr] for attr in ("config_version", "name")
    )


@pytest.mark.unit
def test_resolve_config_using_acceptable_arguments_with_runtime_overrides(
    profiler_with_placeholder_args: RuleBasedProfiler,
) -> None:
    runtime_override_rule_name: str = "my_runtime_override_rule"
    assert all(
        rule.name != runtime_override_rule_name
        for rule in profiler_with_placeholder_args.rules
    )

    runtime_override_rule: dict = {
        "domain_builder": {
            "class_name": "TableDomainBuilder",
            "module_name": "great_expectations.rule_based_profiler.domain_builder",
        },
        "parameter_builders": [
            {
                "class_name": "MetricMultiBatchParameterBuilder",
                "module_name": "great_expectations.rule_based_profiler.parameter_builder",
                "metric_name": "my_other_metric",
                "name": "my_additional_parameter",
            },
        ],
        "expectation_configuration_builders": [
            {
                "class_name": "DefaultExpectationConfigurationBuilder",
                "module_name": "great_expectations.rule_based_profiler.expectation_configuration_builder",
                "expectation_type": "expect_column_values_to_be_between",
                "meta": {
                    "details": {
                        "note": "Here's another rule",
                    },
                },
            },
        ],
    }

    runtime_override_rules: Dict[str, dict] = {
        runtime_override_rule_name: runtime_override_rule,
    }

    config: RuleBasedProfilerConfig = (
        RuleBasedProfilerConfig.resolve_config_using_acceptable_arguments(
            profiler=profiler_with_placeholder_args,
            rules=runtime_override_rules,
        )
    )

    assert len(config.rules) == 2 and runtime_override_rule_name in config.rules


@pytest.mark.unit
def test_resolve_config_using_acceptable_arguments_with_runtime_overrides_with_batch_requests(
    profiler_with_placeholder_args: RuleBasedProfiler,
) -> None:
    runtime_override_rule: dict = {
        "domain_builder": {
            "class_name": "TableDomainBuilder",
            "module_name": "great_expectations.rule_based_profiler.domain_builder",
        },
        "parameter_builders": [
            {
                "class_name": "MetricMultiBatchParameterBuilder",
                "module_name": "great_expectations.rule_based_profiler.parameter_builder",
                "metric_name": "my_other_metric",
                "name": "my_additional_parameter",
            },
        ],
        "expectation_configuration_builders": [
            {
                "class_name": "DefaultExpectationConfigurationBuilder",
                "module_name": "great_expectations.rule_based_profiler.expectation_configuration_builder",
                "expectation_type": "expect_column_values_to_be_between",
                "meta": {
                    "details": {
                        "note": "Here's another rule",
                    },
                },
            },
        ],
    }

    runtime_override_rule_name: str = "rule_with_batch_request"
    runtime_override_rules: Dict[str, dict] = {
        runtime_override_rule_name: runtime_override_rule
    }

    config: RuleBasedProfilerConfig = (
        RuleBasedProfilerConfig.resolve_config_using_acceptable_arguments(
            profiler=profiler_with_placeholder_args,
            rules=runtime_override_rules,
        )
    )

    domain_builder: dict = config.rules[runtime_override_rule_name]["domain_builder"]

    assert domain_builder == {
        "class_name": "TableDomainBuilder",
        "module_name": "great_expectations.rule_based_profiler.domain_builder.table_domain_builder",
    }



import logging

from great_expectations.core._docs_decorators import public_api
from great_expectations.expectations.metrics.metric_provider import MetricProvider

logger = logging.getLogger(__name__)


@public_api
class QueryMetricProvider(MetricProvider):
    """Base class for all Query Metrics, which define metrics to construct SQL queries.

     An example of this is `query.table`,
     which takes in a SQL query & target table name, and returns the result of that query.

     In some cases, subclasses of MetricProvider, such as QueryMetricProvider, will already
     have correct values that may simply be inherited by Metric classes.

     ---Documentation---
         - https://docs.greatexpectations.io/docs/guides/expectations/creating_custom_expectations/how_to_create_custom_query_expectations

    Args:
         metric_name (str): A name identifying the metric. Metric Name must be globally unique in
             a great_expectations installation.
         domain_keys (tuple): A tuple of the keys used to determine the domain of the metric.
         value_keys (tuple): A tuple of the keys used to determine the value of the metric.
         query (str): A valid SQL query.
    """

    domain_keys = ("batch_id", "row_condition", "condition_parser")



import glob
import json
import os
from collections import defaultdict

import pytest

from great_expectations.core.expectation_configuration import ExpectationConfiguration
from great_expectations.data_context.util import file_relative_path
from great_expectations.render.renderer.content_block import (
    ExpectationSuiteBulletListContentBlockRenderer,
)
from great_expectations.render.util import (
    parse_row_condition_string_pandas_engine,
    substitute_none_for_missing,
)


def test_substitute_none_for_missing():
    assert substitute_none_for_missing(
        kwargs={"a": 1, "b": 2}, kwarg_list=["c", "d"]
    ) == {"a": 1, "b": 2, "c": None, "d": None}

    my_kwargs = {"a": 1, "b": 2}
    assert substitute_none_for_missing(kwargs=my_kwargs, kwarg_list=["c", "d"]) == {
        "a": 1,
        "b": 2,
        "c": None,
        "d": None,
    }
    assert my_kwargs == {
        "a": 1,
        "b": 2,
    }, "substitute_none_for_missing should not change input kwargs in place."


def test_parse_row_condition_string_pandas_engine():
    test_condition_string = ""
    assert parse_row_condition_string_pandas_engine(test_condition_string) == (
        "if $row_condition__0",
        {"row_condition__0": "True"},
    )

    test_condition_string = "Age in [0, 42]"
    assert parse_row_condition_string_pandas_engine(test_condition_string) == (
        "if $row_condition__0",
        {"row_condition__0": "Age in [0, 42]"},
    )

    test_condition_string = "Survived == 1 and (SexCode not in (0, 7, x) | ~(Age > 50)) & not (PClass != '1st')"
    assert parse_row_condition_string_pandas_engine(test_condition_string) == (
        "if $row_condition__0 and ($row_condition__1 or not ($row_condition__2)) and not ($row_condition__3)",
        {
            "row_condition__0": "Survived == 1",
            "row_condition__1": "SexCode not in [0, 7, x]",
            "row_condition__2": "Age > 50",
            "row_condition__3": "PClass != '1st'",
        },
    )


@pytest.mark.smoketest
@pytest.mark.rendered_output
@pytest.mark.slow  # 1.52s
def test_all_expectations_using_test_definitions():
    # Chetan - 20220129 - During v0.14.4, it was revealed that this test was broken.
    # The `glob` statement did not pick up any relevant tests, resulting in `test_files` being empty.
    # Without anything to iterate over, the test gave the impression of passing (when in actuality, it never tested anything).
    #
    # After some research, it seems as though this has been broken since the v0.13.0 release.
    # The 5 Expectations noted below are implemented or updated after v0.13.0 and are incompatible with this test fixture due to
    # having incomplete render methods.
    #
    # As this behavior is implemented, the `UNSUPPORTED_EXPECTATIONS` list will be updated to reflect GX's current capabilities.

    dir_path = os.path.dirname(os.path.abspath(__file__))
    pattern = os.path.join(
        dir_path, "..", "..", "tests", "test_definitions", "*", "expect*.json"
    )
    test_files = glob.glob(pattern)

    # Historically, collecting all the JSON tests was an issue - this step ensures we actually have test data.
    assert (
        len(test_files) == 61
    ), "Something went wrong when collecting JSON Expectation test fixtures"

    # The following do not work with this parameterized test due to incomplete render methods.
    UNSUPPORTED_EXPECTATIONS = {
        "expect_column_values_to_match_like_pattern",
        "expect_column_values_to_match_like_pattern_list",
        "expect_column_values_to_not_match_like_pattern",
        "expect_column_values_to_not_match_like_pattern_list",
        "expect_multicolumn_sum_to_equal",
    }

    # Loop over all test_files, datasets, and tests:
    test_results = defaultdict(list)
    for filename in test_files:
        test_definitions = json.load(open(filename))

        # Chetan -20220129 - To be removed once all expectations are supported
        if test_definitions["expectation_type"] in UNSUPPORTED_EXPECTATIONS:
            continue

        for dataset in test_definitions["datasets"]:
            for test in dataset["tests"]:
                # Construct an expectation from the test.
                if type(test["in"]) == dict:
                    fake_expectation = ExpectationConfiguration(
                        expectation_type=test_definitions["expectation_type"],
                        kwargs=test["in"],
                    )
                else:
                    # This would be a good place to put a kwarg-to-arg converter
                    continue

                # Attempt to render it
                print(fake_expectation)
                render_result = ExpectationSuiteBulletListContentBlockRenderer.render(
                    [fake_expectation]
                )
                assert render_result is not None
                render_result = render_result.to_json_dict()

                assert isinstance(render_result, dict)
                assert "content_block_type" in render_result
                assert render_result["content_block_type"] in render_result
                assert isinstance(
                    render_result[render_result["content_block_type"]], list
                )

                # TODO: Assert that the template is renderable, with all the right arguments, etc.
                # rendered_template = pTemplate(el["template"]).substitute(el["params"])

                test_results[test_definitions["expectation_type"]].append(
                    {
                        test["title"]: render_result,
                        # "rendered_template":rendered_template
                    }
                )

    # TODO: accommodate case where multiple datasets exist within one expectation test definition

    # We encountered unicode coding errors on Python 2, but since this is just a smoke test, review the smoke test results in python 3.

    with open(
        file_relative_path(
            __file__,
            "./output/test_render_bullet_list_content_block.json",
            strict=False,
        ),
        "w",
    ) as f:
        json.dump(test_results, f, indent=2)



### Universal datasource configuration elements:

CONNECTION_STRING = "sqlite:///data/yellow_tripdata_sample_2020_all_months_combined.db"


def get_partial_config_universal_datasource_config_elements() -> dict:
    """Creates a dictionary containing the keys and values that are universally defined in
    Spark, Pandas, and SQL Datasource configurations.

    Returns:
         a dictionary containing a partial configuration for a Datasource
    """
    # <snippet name="testerific">
    datasource_config: dict = {
        # <snippet name="test">
        "name": "my_datasource_name",
        "class_name": "Datasource",
        "module_name": "great_expectations.datasource",
        # </snippet>
    }
    # </snippet>
    return datasource_config


### Spark datasource configurations:


def get_full_config_spark_inferred_datasource_single_batch() -> dict:
    """Creates a dictionary configuration for a spark Datasource using an
     inferred data connector that only returns single item batches.

    Returns:
         a dictionary containing a full configuration for a Spark Datasource
    """
    datasource_config: dict = {
        "name": "my_datasource_name",  # Preferably name it something relevant
        "class_name": "Datasource",
        "module_name": "great_expectations.datasource",
        "execution_engine": {
            "class_name": "SparkDFExecutionEngine",
            "module_name": "great_expectations.execution_engine",
        },
        "data_connectors": {
            "name_of_my_inferred_data_connector": {
                "class_name": "InferredAssetFilesystemDataConnector",
                "base_directory": "../data",
                "default_regex": {
                    "pattern": "(.*)\\.csv",
                    "group_names": ["data_asset_name"],
                },
                "batch_spec_passthrough": {
                    "reader_method": "csv",
                    "reader_options": {
                        "header": True,
                        "inferSchema": True,
                    },
                },
            }
        },
    }
    return datasource_config


def get_full_config_spark_inferred_datasource_multi_batch() -> dict:
    """Creates a dictionary configuration for a spark Datasource using an
     inferred data connector that can returns multiple item batches.

    Returns:
         a dictionary containing a full configuration for a Spark Datasource
    """
    datasource_config: dict = {
        "name": "my_datasource_name",  # Preferably name it something relevant
        "class_name": "Datasource",
        "module_name": "great_expectations.datasource",
        "execution_engine": {
            "class_name": "SparkDFExecutionEngine",
            "module_name": "great_expectations.execution_engine",
        },
        "data_connectors": {
            "name_of_my_inferred_data_connector": {
                "class_name": "InferredAssetFilesystemDataConnector",
                "base_directory": "../data",
                "default_regex": {
                    "pattern": "(yellow_tripdata_sample_2020)-(\\d.*)\\.csv",
                    "group_names": ["data_asset_name", "month"],
                },
                "batch_spec_passthrough": {
                    "reader_method": "csv",
                    "reader_options": {
                        "header": True,
                        "inferSchema": True,
                    },
                },
            }
        },
    }
    return datasource_config


def get_full_config_spark_configured_datasource_single_batch() -> dict:
    """Creates a dictionary configuration for a spark Datasource using an
     inferred data connector that only returns single item batches.

    Returns:
         a dictionary containing a full configuration for a Spark Datasource
    """
    datasource_config: dict = {
        "name": "my_datasource_name",  # Preferably name it something relevant
        "class_name": "Datasource",
        "module_name": "great_expectations.datasource",
        "execution_engine": {
            "class_name": "SparkDFExecutionEngine",
            "module_name": "great_expectations.execution_engine",
        },
        "data_connectors": {
            "name_of_my_configured_data_connector": {
                "class_name": "ConfiguredAssetFilesystemDataConnector",
                "base_directory": "../data",
                "assets": {
                    "yellow_tripdata_jan": {
                        "pattern": "yellow_tripdata_sample_2020-(01)\\.csv",
                        "group_names": ["month"],
                    }
                },
                "batch_spec_passthrough": {
                    "reader_method": "csv",
                    "reader_options": {
                        "header": True,
                        "inferSchema": True,
                    },
                },
            }
        },
    }
    return datasource_config


def get_full_config_spark_configured_datasource_multi_batch() -> dict:
    """Creates a dictionary configuration for a spark Datasource using a
     configured data connector that can return multiple item batches.

    Returns:
         a dictionary containing a full configuration for a Spark Datasource
    """
    datasource_config: dict = {
        "name": "my_datasource_name",  # Preferably name it something relevant
        "class_name": "Datasource",
        "module_name": "great_expectations.datasource",
        "execution_engine": {
            "class_name": "SparkDFExecutionEngine",
            "module_name": "great_expectations.execution_engine",
        },
        "data_connectors": {
            "name_of_my_configured_data_connector": {
                "class_name": "ConfiguredAssetFilesystemDataConnector",
                "base_directory": "../data",
                "assets": {
                    "yellow_tripdata_2020": {
                        "pattern": "yellow_tripdata_sample_2020-(.*)\\.csv",
                        "group_names": ["month"],
                    }
                },
                "batch_spec_passthrough": {
                    "reader_method": "csv",
                    "reader_options": {
                        "header": True,
                        "inferSchema": True,
                    },
                },
            }
        },
    }
    return datasource_config


def get_full_config_spark_runtime_datasource() -> dict:
    """Creates a dictionary configuration for a spark Datasource using a
     runtime data connector.

    Returns:
         a dictionary containing a full configuration for a Spark Datasource
    """
    datasource_config: dict = {
        "name": "my_datasource_name",  # Preferably name it something relevant
        "class_name": "Datasource",
        "module_name": "great_expectations.datasource",
        "execution_engine": {
            "class_name": "SparkDFExecutionEngine",
            "module_name": "great_expectations.execution_engine",
        },
        "data_connectors": {
            "name_of_my_runtime_data_connector": {
                "class_name": "RuntimeDataConnector",
                "batch_spec_passthrough": {
                    "reader_method": "csv",
                    "reader_options": {
                        "header": True,
                        "inferSchema": True,
                    },
                },
                "batch_identifiers": ["batch_timestamp"],
            }
        },
    }
    return datasource_config


def get_full_config_sql_inferred_datasource__single_batch_only() -> dict:
    """Creates a dictionary configuration for a SQL Datasource using a
     configured data connector that can only return a single batch.

    Returns:
         a dictionary containing a full configuration for a SQL Datasource
    """
    datasource_config: dict = {
        "name": "my_datasource_name",
        "class_name": "Datasource",
        "module_name": "great_expectations.datasource",
        "execution_engine": {
            "class_name": "SqlAlchemyExecutionEngine",
            "module_name": "great_expectations.execution_engine",
            "connection_string": CONNECTION_STRING,
        },
        "data_connectors": {
            "name_of_my_inferred_data_connector": {
                "class_name": "InferredAssetSqlDataConnector",
            },
        },
    }
    return datasource_config


def get_full_config_sql_inferred_datasource__single_and_multi_batch() -> dict:
    """Creates a dictionary configuration for a SQL Datasource using a
     configured data connector that can return multiple item batches and
     a configured data connector that can only return a single batch.

    Returns:
         a dictionary containing a full configuration for a SQL Datasource
    """
    datasource_config: dict = {
        "name": "my_datasource_name",
        "class_name": "Datasource",
        "module_name": "great_expectations.datasource",
        "execution_engine": {
            "class_name": "SqlAlchemyExecutionEngine",
            "module_name": "great_expectations.execution_engine",
            "connection_string": CONNECTION_STRING,
        },
        "data_connectors": {
            "inferred_data_connector_single_batch_asset": {
                "class_name": "InferredAssetSqlDataConnector",
            },
            "inferred_data_connector_multi_batch_asset_split_on_date_time": {
                "class_name": "InferredAssetSqlDataConnector",
                "splitter_method": "split_on_year_and_month",
                "splitter_kwargs": {
                    "column_name": "pickup_datetime",
                },
            },
        },
    }
    return datasource_config


def get_full_config_sql_configured_datasource() -> dict:
    """Creates a dictionary configuration for a SQL Datasource using a
     inferred data connector that can return multiple item batches and
     a configured data connector that can only return a single batch.

    Returns:
         a dictionary containing a full configuration for a SQL Datasource
    """
    datasource_config: dict = {
        "name": "my_datasource_name",
        "class_name": "Datasource",
        "module_name": "great_expectations.datasource",
        "execution_engine": {
            "class_name": "SqlAlchemyExecutionEngine",
            "module_name": "great_expectations.execution_engine",
            "connection_string": CONNECTION_STRING,
        },
        "data_connectors": {
            "name_of_my_configured_data_connector": {
                "class_name": "ConfiguredAssetSqlDataConnector",
                "assets": {
                    "yellow_tripdata_sample_2020_full": {
                        "table_name": "yellow_tripdata_sample_2020",
                        "schema_name": "main",
                    },
                    "yellow_tripdata_sample_2020_by_year_and_month": {
                        "table_name": "yellow_tripdata_sample_2020",
                        "schema_name": "main",
                        "splitter_method": "split_on_year_and_month",
                        "splitter_kwargs": {
                            "column_name": "pickup_datetime",
                        },
                    },
                },
            },
        },
    }
    return datasource_config


def get_full_config_sql_runtime_datasource() -> dict:
    """Creates a dictionary configuration for a SQL Datasource using a
     runtime data connector.

    Returns:
         a dictionary containing a full configuration for a SQL Datasource
    """
    datasource_config: dict = {
        "name": "my_datasource_name",
        "class_name": "Datasource",
        "module_name": "great_expectations.datasource",
        "execution_engine": {
            "class_name": "SqlAlchemyExecutionEngine",
            "module_name": "great_expectations.execution_engine",
            "connection_string": CONNECTION_STRING,
        },
        "data_connectors": {
            "name_of_my_runtime_data_connector": {
                "class_name": "RuntimeDataConnector",
                "batch_identifiers": ["batch_timestamp"],
            }
        },
    }
    return datasource_config



from __future__ import annotations

from typing import Mapping
from unittest import mock

import pytest

import great_expectations.exceptions as gx_exceptions
from great_expectations.checkpoint.checkpoint import Checkpoint
from great_expectations.core.expectation_configuration import ExpectationConfiguration
from great_expectations.core.expectation_suite import ExpectationSuite
from great_expectations.core.serializer import DictConfigSerializer
from great_expectations.data_context.data_context.ephemeral_data_context import (
    EphemeralDataContext,
)
from great_expectations.data_context.store import ExpectationsStore, ProfilerStore
from great_expectations.data_context.store.checkpoint_store import CheckpointStore
from great_expectations.data_context.store.datasource_store import DatasourceStore
from great_expectations.data_context.types.base import (
    CheckpointConfig,
    DataContextConfig,
    DatasourceConfig,
    InMemoryStoreBackendDefaults,
    ProgressBarsConfig,
    datasourceConfigSchema,
)
from great_expectations.datasource.new_datasource import Datasource
from great_expectations.exceptions.exceptions import StoreConfigurationError
from great_expectations.rule_based_profiler.rule_based_profiler import RuleBasedProfiler


class DatasourceStoreSpy(DatasourceStore):
    def __init__(self) -> None:
        self.save_count = 0
        super().__init__(serializer=DictConfigSerializer(schema=datasourceConfigSchema))

    def set(self, key, value, **kwargs):
        ret = super().set(key=key, value=value, **kwargs)
        self.save_count += 1
        return ret


class ExpectationsStoreSpy(ExpectationsStore):
    def __init__(self) -> None:
        self.save_count = 0
        super().__init__()

    def add(self, key, value, **kwargs):
        ret = super().add(key=key, value=value, **kwargs)
        self.save_count += 1
        return ret

    def update(self, key, value, **kwargs):
        ret = super().update(key=key, value=value, **kwargs)
        self.save_count += 1
        return ret

    def add_or_update(self, key, value, **kwargs):
        ret = super().add_or_update(key=key, value=value, **kwargs)
        self.save_count += 1
        return ret


class ProfilerStoreSpy(ProfilerStore):

    STORE_NAME = "profiler_store"

    def __init__(self) -> None:
        self.save_count = 0
        super().__init__(ProfilerStoreSpy.STORE_NAME)

    def add(self, key, value, **kwargs):
        ret = super().add(key=key, value=value, **kwargs)
        self.save_count += 1
        return ret

    def update(self, key, value, **kwargs):
        ret = super().update(key=key, value=value, **kwargs)
        self.save_count += 1
        return ret

    def add_or_update(self, key, value, **kwargs):
        ret = super().add_or_update(key=key, value=value, **kwargs)
        self.save_count += 1
        return ret


class CheckpointStoreSpy(CheckpointStore):

    STORE_NAME = "checkpoint_store"

    def __init__(self) -> None:
        self.save_count = 0
        super().__init__(CheckpointStoreSpy.STORE_NAME)

    def add(self, key, value, **kwargs):
        ret = super().add(key=key, value=value, **kwargs)
        self.save_count += 1
        return ret

    def update(self, key, value, **kwargs):
        ret = super().update(key=key, value=value, **kwargs)
        self.save_count += 1
        return ret

    def add_or_update(self, key, value, **kwargs):
        ret = super().add_or_update(key=key, value=value, **kwargs)
        self.save_count += 1
        return ret


class EphemeralDataContextSpy(EphemeralDataContext):
    """
    Simply wraps around EphemeralDataContext but keeps tabs on specific method calls around state management.
    """

    def __init__(
        self,
        project_config: DataContextConfig,
    ) -> None:
        super().__init__(project_config)
        self.save_count = 0
        self._datasource_store = DatasourceStoreSpy()
        self._expectations_store = ExpectationsStoreSpy()
        self._profiler_store = ProfilerStoreSpy()
        self._checkpoint_store = CheckpointStoreSpy()

    @property
    def datasource_store(self):
        return self._datasource_store

    @property
    def expectations_store(self):
        return self._expectations_store

    @property
    def profiler_store(self):
        return self._profiler_store

    @property
    def checkpoint_store(self):
        return self._checkpoint_store

    def _save_project_config(self):
        """
        No-op our persistence mechanism but increment an internal counter to ensure it was used.
        """
        self.save_count += 1


@pytest.fixture
def datasource_name() -> str:
    return "my_pandas_datasource"


@pytest.fixture
def in_memory_data_context(
    datasource_name: str,
    datasource_config: DatasourceConfig,
) -> EphemeralDataContextSpy:
    datasources = {
        datasource_name: datasource_config,
    }
    config = DataContextConfig(
        datasources=datasources, store_backend_defaults=InMemoryStoreBackendDefaults()
    )
    context = EphemeralDataContextSpy(project_config=config)
    return context


@pytest.mark.unit
def test_add_store(in_memory_data_context: EphemeralDataContextSpy):
    context = in_memory_data_context

    num_stores_before = len(context.stores)
    num_store_configs_before = len(context.config.stores)

    context.add_store(
        store_name="my_new_store",
        store_config={
            "module_name": "great_expectations.data_context.store",
            "class_name": "ExpectationsStore",
        },
    )

    num_stores_after = len(context.stores)
    num_store_configs_after = len(context.config.stores)

    assert num_stores_after == num_stores_before + 1
    assert num_store_configs_after == num_store_configs_before + 1
    assert context.save_count == 1


@pytest.mark.unit
def test_delete_store_success(in_memory_data_context: EphemeralDataContextSpy):
    context = in_memory_data_context

    num_stores_before = len(context.stores)
    num_store_configs_before = len(context.config.stores)

    context.delete_store("checkpoint_store")  # We know this to be a default name

    num_stores_after = len(context.stores)
    num_store_configs_after = len(context.config.stores)

    assert num_stores_after == num_stores_before - 1
    assert num_store_configs_after == num_store_configs_before - 1
    assert context.save_count == 1


@pytest.mark.unit
def test_delete_store_failure(in_memory_data_context: EphemeralDataContextSpy):
    context = in_memory_data_context

    num_stores_before = len(context.stores)
    num_store_configs_before = len(context.config.stores)

    with pytest.raises(StoreConfigurationError):
        context.delete_store("my_fake_store_name")

    num_stores_after = len(context.stores)
    num_store_configs_after = len(context.config.stores)

    assert num_stores_after == num_stores_before
    assert num_store_configs_after == num_store_configs_before
    assert context.save_count == 0


@pytest.mark.unit
@pytest.mark.parametrize(
    "config",
    [
        pytest.param(
            DataContextConfig(progress_bars=ProgressBarsConfig(globally=True)),
            id="DataContextConfig",
        ),
        pytest.param(
            {"progress_bars": ProgressBarsConfig(globally=True)}, id="Mapping"
        ),
    ],
)
def test_update_project_config(
    in_memory_data_context: EphemeralDataContextSpy, config: DataContextConfig | Mapping
):
    context = in_memory_data_context

    assert context.progress_bars is None

    context.update_project_config(config)

    assert context.progress_bars["globally"] is True


@pytest.mark.unit
def test_add_datasource_with_existing_datasource(
    in_memory_data_context: EphemeralDataContextSpy,
    datasource_config: DatasourceConfig,
):
    context = in_memory_data_context

    config_dict = datasource_config.to_dict()
    for attr in ("class_name", "module_name"):
        config_dict.pop(attr)
    config_dict["name"] = "my_datasource"

    datasource = Datasource(**config_dict)
    persisted_datasource = context.add_datasource(datasource=datasource)

    expected_config = datasource.config
    actual_config = persisted_datasource.config

    # Name gets injected into data connector as part of serialization hook
    # Removing for purposes of test assertion
    data_connectors = actual_config["data_connectors"]
    data_connector_name = tuple(data_connectors.keys())[0]
    data_connectors[data_connector_name].pop("name")

    assert actual_config == expected_config
    assert context.datasource_store.save_count == 1


@pytest.mark.unit
@pytest.mark.parametrize(
    "datasource,datasource_name",
    [
        pytest.param(
            mock.MagicMock(),
            "my_datasource",
            id="both datasource and name",
        ),
        pytest.param(None, None, id="neither datasource nor name"),
    ],
)
def test_add_datasource_conflicting_args_failure(
    in_memory_data_context: EphemeralDataContextSpy,
    datasource: mock.MagicMock | None,
    datasource_name: str | None,
):
    context = in_memory_data_context

    with pytest.raises(ValueError) as e:
        context.add_datasource(datasource=datasource, name=datasource_name)

    assert (
        "an existing datasource or individual constructor arguments (but not both)"
        in str(e.value)
    )
    assert context.datasource_store.save_count == 0


@pytest.mark.unit
def test_add_or_update_datasource_updates_with_individual_args_successfully(
    in_memory_data_context: EphemeralDataContextSpy,
    datasource_name: str,
    datasource_config: DatasourceConfig,
):
    context = in_memory_data_context

    num_datasource_before = len(context.datasources)
    num_datasource_configs_before = len(context.config.datasources)

    assert (
        datasource_name in context.datasources
    ), f"Downstream logic in the test relies on {datasource_name} being a datasource; please check your fixtures."

    config_dict = datasource_config.to_dict()
    config_dict["name"] = datasource_name

    id = "d53c2384-f973-4a0c-9c85-af1d67c06f58"
    config_dict["id"] = id

    datasource = context.add_or_update_datasource(**config_dict)
    # Let's `id` as an example attr to change so we don't need to assert against the whole config
    assert datasource.config["id"] == id

    num_datasource_after = len(context.datasources)
    num_datasource_configs_after = len(context.config.datasources)

    assert num_datasource_after == num_datasource_before
    assert num_datasource_configs_after == num_datasource_configs_before
    assert context.datasource_store.save_count == 1


@pytest.mark.unit
def test_add_or_update_datasource_updates_with_existing_datasource_successfully(
    in_memory_data_context: EphemeralDataContextSpy,
    datasource_name: str,
    datasource_config: DatasourceConfig,
):
    context = in_memory_data_context

    num_datasource_before = len(context.datasources)
    num_datasource_configs_before = len(context.config.datasources)

    assert (
        datasource_name in context.datasources
    ), f"Downstream logic in the test relies on {datasource_name} being a datasource; please check your fixtures."

    config_dict = datasource_config.to_dict()
    config_dict["name"] = datasource_name
    for attr in ("class_name", "module_name"):
        config_dict.pop(attr)

    datasource = Datasource(**config_dict)
    persisted_datasource = context.add_or_update_datasource(datasource=datasource)

    num_datasource_after = len(context.datasources)
    num_datasource_configs_after = len(context.config.datasources)

    assert num_datasource_after == num_datasource_before
    assert num_datasource_configs_after == num_datasource_configs_before
    assert context.datasource_store.save_count == 1

    expected_config = datasource.config
    actual_config = persisted_datasource.config

    # Name gets injected into data connector as part of serialization hook
    # Removing for purposes of test assertion
    data_connectors = actual_config["data_connectors"]
    data_connector_name = tuple(data_connectors.keys())[0]
    data_connectors[data_connector_name].pop("name")

    assert actual_config == expected_config


@pytest.mark.unit
@pytest.mark.parametrize("use_existing_datasource", [True, False])
def test_add_or_update_datasource_adds_successfully(
    in_memory_data_context: EphemeralDataContextSpy,
    datasource_config: DatasourceConfig,
    use_existing_datasource: bool,
):
    context = in_memory_data_context

    num_datasource_before = len(context.datasources)
    num_datasource_configs_before = len(context.config.datasources)

    datasource_name = "my_brand_new_datasource"

    assert datasource_name not in context.datasources

    config_dict = datasource_config.to_dict()
    config_dict["name"] = datasource_name

    if use_existing_datasource:
        for attr in ("class_name", "module_name"):
            config_dict.pop(attr)
        datasource = Datasource(**config_dict)
        _ = context.add_or_update_datasource(datasource=datasource)
    else:
        _ = context.add_or_update_datasource(**config_dict)

    num_datasource_after = len(context.datasources)
    num_datasource_configs_after = len(context.config.datasources)

    assert datasource_name in context.datasources
    assert num_datasource_after == num_datasource_before + 1
    assert num_datasource_configs_after == num_datasource_configs_before + 1
    assert context.datasource_store.save_count == 1


@pytest.mark.unit
@pytest.mark.parametrize(
    "datasource,datasource_name",
    [
        pytest.param(
            mock.MagicMock(),
            "my_datasource",
            id="both datasource and name",
        ),
        pytest.param(None, None, id="neither datasource nor name"),
    ],
)
def test_add_or_update_datasource_conflicting_args_failure(
    in_memory_data_context: EphemeralDataContextSpy,
    datasource: mock.MagicMock | None,
    datasource_name: str | None,
):
    context = in_memory_data_context

    with pytest.raises(ValueError) as e:
        context.add_or_update_datasource(datasource=datasource, name=datasource_name)

    assert (
        "an existing datasource or individual constructor arguments (but not both)"
        in str(e.value)
    )
    assert context.expectations_store.save_count == 0


@pytest.mark.unit
@pytest.mark.parametrize(
    "kwargs,expected_suite",
    [
        pytest.param(
            {"expectation_suite_name": "my_new_suite"},
            ExpectationSuite(expectation_suite_name="my_new_suite"),
            id="only name",
        ),
        pytest.param(
            {
                "expectations": [
                    ExpectationConfiguration(
                        expectation_type="expect_column_values_to_be_in_set",
                        kwargs={"column": "x", "value_set": [1, 2, 4]},
                    ),
                ],
                "expectation_suite_name": "default",
                "meta": {"great_expectations_version": "0.15.44"},
            },
            ExpectationSuite(
                expectations=[
                    ExpectationConfiguration(
                        expectation_type="expect_column_values_to_be_in_set",
                        kwargs={"column": "x", "value_set": [1, 2, 4]},
                    ),
                ],
                expectation_suite_name="default",
                meta={"great_expectations_version": "0.15.44"},
            ),
            id="misc args",
        ),
        pytest.param(
            {
                "expectation_suite": ExpectationSuite(
                    expectations=[
                        ExpectationConfiguration(
                            expectation_type="expect_column_values_to_be_in_set",
                            kwargs={"column": "x", "value_set": [1, 2, 4]},
                        ),
                    ],
                    expectation_suite_name="default",
                    meta={"great_expectations_version": "0.15.44"},
                ),
            },
            ExpectationSuite(
                expectations=[
                    ExpectationConfiguration(
                        expectation_type="expect_column_values_to_be_in_set",
                        kwargs={"column": "x", "value_set": [1, 2, 4]},
                    ),
                ],
                expectation_suite_name="default",
                meta={"great_expectations_version": "0.15.44"},
            ),
            id="existing suite",
        ),
    ],
)
def test_add_expectation_suite_success(
    in_memory_data_context: EphemeralDataContextSpy,
    kwargs: dict,
    expected_suite: ExpectationSuite,
):
    context = in_memory_data_context

    suite = context.add_expectation_suite(**kwargs)

    assert suite == expected_suite
    assert context.expectations_store.save_count == 1


@pytest.mark.unit
def test_add_expectation_suite_namespace_collision_failure(
    in_memory_data_context: EphemeralDataContextSpy,
):
    context = in_memory_data_context

    suite_name = "default"
    context.add_expectation_suite(expectation_suite_name=suite_name)

    with pytest.raises(gx_exceptions.DataContextError) as e:
        context.add_expectation_suite(expectation_suite_name=suite_name)

    assert f"An ExpectationSuite named {suite_name} already exists" in str(e.value)
    assert context.expectations_store.save_count == 1


@pytest.mark.unit
@pytest.mark.parametrize(
    "suite,suite_name",
    [
        pytest.param(
            ExpectationSuite(expectation_suite_name="default"),
            "default",
            id="both suite and suite_name",
        ),
        pytest.param(None, None, id="neither suite nor suite_name"),
    ],
)
def test_add_expectation_suite_conflicting_args_failure(
    in_memory_data_context: EphemeralDataContextSpy,
    suite: ExpectationSuite | None,
    suite_name: str | None,
):
    context = in_memory_data_context

    with pytest.raises(ValueError) as e:
        context.add_expectation_suite(
            expectation_suite=suite, expectation_suite_name=suite_name
        )

    assert (
        "an existing expectation_suite or individual constructor arguments (but not both)"
        in str(e.value)
    )
    assert context.expectations_store.save_count == 0


@pytest.mark.unit
def test_update_expectation_suite_success(
    in_memory_data_context: EphemeralDataContextSpy,
):
    context = in_memory_data_context

    suite_name = "default"
    suite = context.add_expectation_suite(suite_name)

    assert context.expectations_store.save_count == 1

    suite.expectations = [
        ExpectationConfiguration(
            expectation_type="expect_column_values_to_be_in_set",
            kwargs={"column": "x", "value_set": [1, 2, 4]},
        ),
    ]
    updated_suite = context.update_expectation_suite(suite)

    assert updated_suite.expectations == suite.expectations
    assert context.expectations_store.save_count == 2


@pytest.mark.unit
def test_update_expectation_suite_failure(
    in_memory_data_context: EphemeralDataContextSpy,
):
    context = in_memory_data_context

    suite_name = "my_brand_new_suite"
    suite = ExpectationSuite(expectation_suite_name=suite_name)

    with pytest.raises(gx_exceptions.ExpectationSuiteError) as e:
        _ = context.update_expectation_suite(suite)

    assert f"Could not find an existing ExpectationSuite named {suite_name}." in str(
        e.value
    )


@pytest.mark.unit
@pytest.mark.parametrize(
    "kwargs",
    [
        pytest.param(
            {
                "expectation_suite_name": "default",
                "expectations": [
                    ExpectationConfiguration(
                        expectation_type="expect_column_values_to_be_in_set",
                        kwargs={"column": "x", "value_set": [1, 2, 4]},
                    ),
                ],
                "meta": {"great_expectations_version": "0.15.44"},
            },
            id="individual args",
        ),
        pytest.param(
            {
                "expectation_suite": ExpectationSuite(
                    expectations=[
                        ExpectationConfiguration(
                            expectation_type="expect_column_values_to_be_in_set",
                            kwargs={"column": "x", "value_set": [1, 2, 4]},
                        ),
                    ],
                    expectation_suite_name="default",
                    meta={"great_expectations_version": "0.15.44"},
                ),
            },
            id="existing suite",
        ),
    ],
)
def test_add_or_update_expectation_suite_adds_successfully(
    in_memory_data_context: EphemeralDataContextSpy,
    kwargs: dict,
):
    context = in_memory_data_context

    expectation_suite_name = "default"
    expectations = [
        ExpectationConfiguration(
            expectation_type="expect_column_values_to_be_in_set",
            kwargs={"column": "x", "value_set": [1, 2, 4]},
        ),
    ]
    meta = {"great_expectations_version": "0.15.44"}

    suite = context.add_or_update_expectation_suite(**kwargs)

    assert suite.expectation_suite_name == expectation_suite_name
    assert suite.expectations == expectations
    assert suite.meta == meta
    assert context.expectations_store.save_count == 1


@pytest.mark.unit
@pytest.mark.parametrize(
    "new_expectations",
    [
        pytest.param(
            [],
            id="empty expectations",
        ),
        pytest.param(
            [
                ExpectationConfiguration(
                    expectation_type="expect_column_to_exist",
                    kwargs={"column": "pickup_datetime"},
                )
            ],
            id="new expectations",
        ),
    ],
)
def test_add_or_update_expectation_suite_updates_successfully(
    in_memory_data_context: EphemeralDataContextSpy,
    new_expectations: list[ExpectationConfiguration],
):
    context = in_memory_data_context

    suite_name = "default"
    suite = context.add_expectation_suite(
        expectation_suite_name=suite_name,
        expectations=[
            ExpectationConfiguration(
                expectation_type="expect_column_values_to_be_in_set",
                kwargs={"column": "x", "value_set": [1, 2, 4]},
            ),
        ],
    )

    assert context.expectations_store.save_count == 1

    suite = context.add_or_update_expectation_suite(
        expectation_suite_name=suite_name, expectations=new_expectations
    )

    assert suite.expectations == new_expectations
    assert context.expectations_store.save_count == 2


@pytest.mark.unit
@pytest.mark.parametrize(
    "suite,suite_name",
    [
        pytest.param(
            ExpectationSuite(expectation_suite_name="default"),
            "default",
            id="both suite and suite_name",
        ),
        pytest.param(None, None, id="neither suite nor suite_name"),
    ],
)
def test_add_or_update_expectation_suite_conflicting_args_failure(
    in_memory_data_context: EphemeralDataContextSpy,
    suite: ExpectationSuite | None,
    suite_name: str | None,
):
    context = in_memory_data_context

    with pytest.raises((ValueError, AssertionError)) as e:
        context.add_or_update_expectation_suite(
            expectation_suite=suite, expectation_suite_name=suite_name
        )

    assert context.expectations_store.save_count == 0


@pytest.mark.unit
def test_add_profiler_with_existing_profiler(
    in_memory_data_context: EphemeralDataContextSpy,
    profiler_rules: dict,
):
    context = in_memory_data_context

    name = "my_rbp"
    profiler = RuleBasedProfiler(
        name=name,
        config_version=1.0,
        rules=profiler_rules,
        data_context=context,
    )

    persisted_profiler = context.add_profiler(profiler=profiler)

    assert profiler.name == persisted_profiler.name
    assert profiler.config_version == persisted_profiler.config_version
    assert len(profiler.rules) == len(persisted_profiler.rules)
    assert context.profiler_store.save_count == 1


@pytest.mark.unit
def test_add_profiler_namespace_collision_raises_deprecation(
    in_memory_data_context: EphemeralDataContextSpy,
    profiler_rules: dict,
):
    context = in_memory_data_context

    name = "my_rbp"
    profiler = RuleBasedProfiler(
        name=name,
        config_version=1.0,
        rules=profiler_rules,
        data_context=context,
    )

    _ = context.add_profiler(profiler=profiler)
    assert context.profiler_store.save_count == 1

    with pytest.deprecated_call():
        _ = context.add_profiler(profiler=profiler)

    assert context.profiler_store.save_count == 2


@pytest.mark.unit
@pytest.mark.parametrize(
    "profiler,profiler_name",
    [
        pytest.param(
            mock.MagicMock(),
            "my_rbp",
            id="both profiler and profiler_name",
        ),
        pytest.param(None, None, id="neither profiler nor profiler_name"),
    ],
)
def test_add_profiler_conflicting_args_failure(
    in_memory_data_context: EphemeralDataContextSpy,
    profiler: mock.MagicMock | None,
    profiler_name: str | None,
):
    context = in_memory_data_context

    with pytest.raises(ValueError) as e:
        context.add_profiler(profiler=profiler, name=profiler_name)

    assert (
        "an existing profiler or individual constructor arguments (but not both)"
        in str(e.value)
    )
    assert context.profiler_store.save_count == 0


@pytest.mark.unit
def test_update_profiler_success(
    in_memory_data_context: EphemeralDataContextSpy,
    profiler_rules: dict,
):
    context = in_memory_data_context

    name = "my_rbp"
    config_version = 1.0
    rules = profiler_rules

    profiler = context.add_profiler(
        name=name, config_version=config_version, rules=rules
    )

    assert context.profiler_store.save_count == 1

    profiler.rules = []
    updated_profiler = context.update_profiler(profiler)

    assert updated_profiler.rules == profiler.rules
    assert context.profiler_store.save_count == 2


@pytest.mark.unit
def test_update_profiler_failure(
    in_memory_data_context: EphemeralDataContextSpy, profiler_rules: dict
):
    context = in_memory_data_context

    name = "my_rbp"
    profiler = RuleBasedProfiler(
        name=name,
        config_version=1.0,
        rules=profiler_rules,
        data_context=context,
    )

    with pytest.raises(gx_exceptions.ProfilerNotFoundError) as e:
        _ = context.update_profiler(profiler)

    assert f"Could not find an existing Profiler named {name}" in str(e.value)


@pytest.mark.unit
def test_add_or_update_profiler_adds_successfully(
    in_memory_data_context: EphemeralDataContextSpy, profiler_rules: dict
):
    context = in_memory_data_context

    name = "my_rbp"
    config_version = 1.0
    rules = profiler_rules

    profiler = context.add_or_update_profiler(
        name=name, config_version=config_version, rules=rules
    )

    config = profiler.config

    assert config.name == name
    assert config.config_version == config_version
    assert len(config.rules) == len(rules) and config.rules.keys() == rules.keys()
    assert context.profiler_store.save_count == 1


@pytest.mark.unit
@pytest.mark.parametrize(
    "new_rules",
    [
        pytest.param(
            {},
            id="empty rules",
        ),
        pytest.param(
            {
                "my_new_rule": {
                    "variables": {},
                    "domain_builder": {
                        "class_name": "TableDomainBuilder",
                    },
                    "parameter_builders": [
                        {
                            "class_name": "MetricMultiBatchParameterBuilder",
                            "name": "my_parameter",
                            "metric_name": "my_metric",
                        },
                    ],
                    "expectation_configuration_builders": [
                        {
                            "class_name": "DefaultExpectationConfigurationBuilder",
                            "expectation_type": "expect_column_pair_values_A_to_be_greater_than_B",
                            "column_A": "$domain.domain_kwargs.column_A",
                            "column_B": "$domain.domain_kwargs.column_B",
                        },
                    ],
                },
            },
            id="new rule",
        ),
    ],
)
def test_add_or_update_profiler_updates_successfully(
    in_memory_data_context: EphemeralDataContextSpy,
    profiler_rules: dict,
    new_rules: dict,
):
    context = in_memory_data_context

    name = "my_rbp"
    config_version = 1.0
    rules = profiler_rules

    profiler = context.add_profiler(
        name=name, config_version=config_version, rules=rules
    )

    assert context.profiler_store.save_count == 1

    profiler = context.add_or_update_profiler(
        name=name, rules=new_rules, config_version=config_version
    )

    # Rules get converted to a list within the RBP constructor
    assert sorted(rule.name for rule in profiler.rules) == sorted(new_rules.keys())
    assert context.profiler_store.save_count == 2


@pytest.mark.unit
@pytest.mark.parametrize(
    "profiler,profiler_name",
    [
        pytest.param(
            mock.MagicMock(),  # Only care about the presence of the value (no need to construct a full RBP obj)
            "my_rbp",
            id="both profiler and profiler_name",
        ),
        pytest.param(None, None, id="neither profiler nor profiler_name"),
    ],
)
def test_add_or_update_profiler_conflicting_args_failure(
    in_memory_data_context: EphemeralDataContextSpy,
    profiler: mock.MagicMock | None,
    profiler_name: str | None,
):
    context = in_memory_data_context

    with pytest.raises(ValueError) as e:
        context.add_or_update_profiler(profiler=profiler, name=profiler_name)

    assert (
        "an existing profiler or individual constructor arguments (but not both)"
        in str(e.value)
    )
    assert context.profiler_store.save_count == 0


@pytest.mark.unit
def test_add_checkpoint_with_existing_checkpoint(
    in_memory_data_context: EphemeralDataContextSpy,
):
    context = in_memory_data_context
    checkpoint_name = "my_checkpoint"
    checkpoint = Checkpoint(name=checkpoint_name, data_context=context)

    persisted_checkpoint = context.add_checkpoint(checkpoint=checkpoint)

    assert checkpoint == persisted_checkpoint
    assert context.checkpoint_store.save_count == 1


@pytest.mark.unit
def test_add_checkpoint_namespace_collision_raises_deprecation(
    in_memory_data_context: EphemeralDataContextSpy,
):
    context = in_memory_data_context
    checkpoint_name = "my_checkpoint"

    _ = context.add_checkpoint(name=checkpoint_name, class_name="Checkpoint")
    assert context.checkpoint_store.save_count == 1

    with pytest.deprecated_call():
        _ = context.add_checkpoint(name=checkpoint_name, class_name="Checkpoint")

    assert context.checkpoint_store.save_count == 2


@pytest.mark.unit
@pytest.mark.parametrize(
    "checkpoint, checkpoint_name",
    [
        pytest.param(
            mock.MagicMock(),
            "my_checkpoint_name",
            id="both checkpoint and checkpoint_name",
        ),
        pytest.param(None, None, id="neither checkpoint nor checkpoint_name"),
    ],
)
def test_add_checkpoint_conflicting_args_failure(
    in_memory_data_context: EphemeralDataContextSpy,
    # Only care about the presence of the value (no need to construct a full Checkpoint obj)
    checkpoint: mock.MagicMock | None,
    checkpoint_name: str | None,
):
    context = in_memory_data_context

    with pytest.raises(ValueError) as e:
        context.add_checkpoint(
            checkpoint=checkpoint,
            name=checkpoint_name,
        )

    assert (
        "an existing checkpoint or individual constructor arguments (but not both)"
        in str(e.value)
    )
    assert context.checkpoint_store.save_count == 0


@pytest.mark.unit
def test_update_checkpoint_success(
    in_memory_data_context: EphemeralDataContextSpy,
    checkpoint_config: dict,
):
    context = in_memory_data_context
    name = "my_checkpoint"
    checkpoint_config["name"] = name

    checkpoint = context.add_checkpoint(**checkpoint_config)

    assert context.checkpoint_store.save_count == 1

    action_list = [
        {
            "name": "store_validation_result",
            "action": {
                "class_name": "StoreValidationResultAction",
            },
        },
        {
            "name": "store_evaluation_params",
            "action": {
                "class_name": "StoreEvaluationParametersAction",
            },
        },
        {
            "name": "update_data_docs",
            "action": {
                "class_name": "UpdateDataDocsAction",
            },
        },
    ]
    checkpoint.config.action_list = action_list
    context.update_checkpoint(checkpoint)

    assert context.checkpoint_store.save_count == 2


@pytest.mark.unit
def test_update_checkpoint_failure(in_memory_data_context: EphemeralDataContextSpy):
    context = in_memory_data_context

    name = "my_checkpoint"
    checkpoint = Checkpoint(
        name=name,
        data_context=context,
    )

    with pytest.raises(gx_exceptions.CheckpointNotFoundError) as e:
        context.update_checkpoint(checkpoint)

    assert f"Could not find an existing Checkpoint named {name}" in str(e.value)


@pytest.mark.unit
@pytest.mark.parametrize("use_existing_checkpoint", [True, False])
def test_add_or_update_checkpoint_adds_successfully(
    in_memory_data_context: EphemeralDataContextSpy,
    checkpoint_config: dict,
    use_existing_checkpoint: bool,
):
    context = in_memory_data_context

    if use_existing_checkpoint:
        checkpoint_config.pop("class_name")
        checkpoint = Checkpoint(**checkpoint_config, data_context=context)
        checkpoint = context.add_or_update_checkpoint(checkpoint=checkpoint)
    else:
        checkpoint = context.add_or_update_checkpoint(**checkpoint_config)

    actual_config = checkpoint.config

    assert actual_config.name == checkpoint_config["name"]
    assert (
        actual_config.expectation_suite_name
        == checkpoint_config["expectation_suite_name"]
    )
    assert actual_config.validations == checkpoint_config["validations"]
    assert context.checkpoint_store.save_count == 1


@pytest.mark.unit
@pytest.mark.parametrize(
    "update_kwargs,expected_config",
    [
        pytest.param(
            {
                "action_list": [],
                "batch_request": {},
                "class_name": "Checkpoint",
                "config_version": 1.0,
                "evaluation_parameters": {},
                "expectation_suite_name": "oss_test_expectation_suite",
                "module_name": "great_expectations.checkpoint",
                "profilers": [],
                "runtime_configuration": {},
                "validations": [
                    {"expectation_suite_name": "taxi.demo_pass"},
                    {
                        "batch_request": {
                            "datasource_name": "oss_test_datasource",
                            "data_connector_name": "oss_test_data_connector",
                            "data_asset_name": "users",
                        }
                    },
                ],
            },
            CheckpointConfig(
                **{
                    "action_list": [],
                    "batch_request": {},
                    "class_name": "Checkpoint",
                    "config_version": 1.0,
                    "evaluation_parameters": {},
                    "expectation_suite_name": "oss_test_expectation_suite",
                    "module_name": "great_expectations.checkpoint",
                    "name": "my_checkpoint",
                    "profilers": [],
                    "runtime_configuration": {},
                    "validations": [
                        {"expectation_suite_name": "taxi.demo_pass"},
                        {
                            "batch_request": {
                                "datasource_name": "oss_test_datasource",
                                "data_connector_name": "oss_test_data_connector",
                                "data_asset_name": "users",
                            }
                        },
                    ],
                }
            ),
            id="no updates",
        ),
        pytest.param(
            {
                "class_name": "Checkpoint",
                "module_name": "great_expectations.checkpoint",
                "config_version": 1.0,
                "validations": [],
                "evaluation_parameters": {
                    "environment": "$GE_ENVIRONMENT",
                    "tolerance": 1.0e-2,
                    "aux_param_0": "$MY_PARAM",
                    "aux_param_1": "1 + $MY_PARAM",
                },
            },
            CheckpointConfig(
                **{
                    "action_list": [],
                    "batch_request": {},
                    "class_name": "Checkpoint",
                    "config_version": 1.0,
                    "evaluation_parameters": {
                        "environment": "$GE_ENVIRONMENT",
                        "tolerance": 1.0e-2,
                        "aux_param_0": "$MY_PARAM",
                        "aux_param_1": "1 + $MY_PARAM",
                    },
                    "expectation_suite_name": None,
                    "module_name": "great_expectations.checkpoint",
                    "name": "my_checkpoint",
                    "profilers": [],
                    "runtime_configuration": {},
                    "validations": [],
                }
            ),
            id="some updates",
        ),
    ],
)
def test_add_or_update_checkpoint_individual_args_updates_successfully(
    in_memory_data_context: EphemeralDataContextSpy,
    checkpoint_config: dict,
    update_kwargs: dict,
    expected_config: CheckpointConfig,
):
    context = in_memory_data_context
    name = "my_checkpoint"
    checkpoint_config["name"] = name

    checkpoint = context.add_checkpoint(**checkpoint_config)

    assert context.checkpoint_store.save_count == 1

    checkpoint = context.add_or_update_checkpoint(name=name, **update_kwargs)

    assert checkpoint.config.to_dict() == expected_config.to_dict()
    assert context.checkpoint_store.save_count == 2


@pytest.mark.unit
def test_add_or_update_checkpoint_existing_checkpoint_updates_successfully(
    in_memory_data_context: EphemeralDataContextSpy,
    checkpoint_config: dict,
):
    context = in_memory_data_context

    name = "my_checkpoint"
    checkpoint_config["name"] = name

    checkpoint = context.add_checkpoint(name=name, class_name="Checkpoint")

    assert len(checkpoint.validations) == 0
    assert context.checkpoint_store.save_count == 1

    checkpoint_config.pop("class_name")
    checkpoint = Checkpoint(**checkpoint_config, data_context=context)
    checkpoint = context.add_or_update_checkpoint(checkpoint=checkpoint)

    assert len(checkpoint.validations) == len(checkpoint_config["validations"])
    assert context.checkpoint_store.save_count == 2


@pytest.mark.unit
@pytest.mark.parametrize(
    "checkpoint, checkpoint_name",
    [
        pytest.param(
            mock.MagicMock(),
            "my_checkpoint_name",
            id="both checkpoint and checkpoint_name",
        ),
        pytest.param(None, None, id="neither checkpoint nor checkpoint_name"),
    ],
)
def test_add_or_update_checkpoint_conflicting_args_failure(
    in_memory_data_context: EphemeralDataContextSpy,
    # Only care about the presence of the value (no need to construct a full Checkpoint obj)
    checkpoint: mock.MagicMock | None,
    checkpoint_name: str | None,
):
    context = in_memory_data_context

    with pytest.raises(ValueError) as e:
        context.add_or_update_checkpoint(
            checkpoint=checkpoint,
            name=checkpoint_name,
        )

    assert (
        "an existing checkpoint or individual constructor arguments (but not both)"
        in str(e.value)
    )
    assert context.checkpoint_store.save_count == 0



from dataclasses import make_dataclass
from typing import Any, Dict, List, Set, Tuple, Union

from great_expectations.validator.computed_metric import MetricValue

MetricValues = Union[
    MetricValue, List[MetricValue], Set[MetricValue], Tuple[MetricValue, ...]
]
MetricComputationDetails = Dict[str, Any]
MetricComputationResult = make_dataclass(
    "MetricComputationResult", ["attributed_resolved_metrics", "details"]
)



from typing import List

from great_expectations import DataContext
from great_expectations.data_context.types.base import DatasourceConfig


def test_list_datasources_base_data_context_no_datasources(
    empty_data_context: DataContext, datasource_config_with_names: DatasourceConfig
) -> None:
    """What does this test and why?

    When listing datasources, we want to omit the name and id fields. This test uses DataContext.
    """

    context: DataContext = empty_data_context

    # no datasources

    observed: List[dict] = context.list_datasources()

    expected: List[dict] = []

    assert observed == expected


def test_list_datasources_base_data_context_one_datasource(
    empty_data_context: DataContext, datasource_config_with_names: DatasourceConfig
) -> None:
    """What does this test and why?

    When listing datasources, we want to omit the name and id fields. This test uses DataContext.
    """

    context: DataContext = empty_data_context

    # one datasource

    context.add_datasource(**datasource_config_with_names.to_dict())

    observed = context.list_datasources()

    expected = [
        {
            "class_name": "Datasource",
            "data_connectors": {
                "tripdata_monthly_configured": {
                    "assets": {
                        "yellow": {
                            "class_name": "Asset",
                            "group_names": ["year", "month"],
                            "module_name": "great_expectations.datasource.data_connector.asset",
                            "pattern": "yellow_tripdata_(\\d{4})-(\\d{2})\\.csv$",
                        }
                    },
                    "base_directory": "/path/to/trip_data",
                    "class_name": "ConfiguredAssetFilesystemDataConnector",
                    "module_name": "great_expectations.datasource.data_connector",
                    # NOTE: no "name" field included for data connectors
                }
            },
            "execution_engine": {
                "class_name": "PandasExecutionEngine",
                "module_name": "great_expectations.execution_engine",
            },
            "module_name": "great_expectations.datasource",
            "name": "my_datasource",
        }
    ]

    assert observed == expected


def test_list_datasources_base_data_context_two_datasources(
    empty_data_context: DataContext, datasource_config_with_names: DatasourceConfig
) -> None:
    """What does this test and why?

    When listing datasources, we want to omit the name and id fields. This test uses DataContext.
    """

    context: DataContext = empty_data_context

    # two datasources

    context.add_datasource(**datasource_config_with_names.to_dict())

    datasource_config_two_with_names = datasource_config_with_names
    datasource_config_two_with_names.name = "datasource_two_name"

    context.add_datasource(**datasource_config_two_with_names.to_dict())

    observed = context.list_datasources()

    expected = [
        {
            "class_name": "Datasource",
            "data_connectors": {
                "tripdata_monthly_configured": {
                    "assets": {
                        "yellow": {
                            "class_name": "Asset",
                            "group_names": ["year", "month"],
                            "module_name": "great_expectations.datasource.data_connector.asset",
                            "pattern": "yellow_tripdata_(\\d{4})-(\\d{2})\\.csv$",
                        }
                    },
                    "base_directory": "/path/to/trip_data",
                    "class_name": "ConfiguredAssetFilesystemDataConnector",
                    "module_name": "great_expectations.datasource.data_connector",
                    # NOTE: no "name" field included for data connectors
                }
            },
            "execution_engine": {
                "class_name": "PandasExecutionEngine",
                "module_name": "great_expectations.execution_engine",
            },
            "module_name": "great_expectations.datasource",
            "name": "my_datasource",
        },
        {
            "class_name": "Datasource",
            "data_connectors": {
                "tripdata_monthly_configured": {
                    "assets": {
                        "yellow": {
                            "class_name": "Asset",
                            "group_names": ["year", "month"],
                            "module_name": "great_expectations.datasource.data_connector.asset",
                            "pattern": "yellow_tripdata_(\\d{4})-(\\d{2})\\.csv$",
                        }
                    },
                    "base_directory": "/path/to/trip_data",
                    "class_name": "ConfiguredAssetFilesystemDataConnector",
                    "module_name": "great_expectations.datasource.data_connector",
                    # NOTE: no "name" field included for data connectors
                }
            },
            "execution_engine": {
                "class_name": "PandasExecutionEngine",
                "module_name": "great_expectations.execution_engine",
            },
            "module_name": "great_expectations.datasource",
            "name": "datasource_two_name",
        },
    ]

    assert observed == expected



from typing import Dict, Optional

from great_expectations.core.expectation_configuration import ExpectationConfiguration
from great_expectations.core.metric_domain_types import MetricDomainTypes
from great_expectations.core.metric_function_types import MetricPartialFunctionTypes
from great_expectations.execution_engine import (
    ExecutionEngine,
    PandasExecutionEngine,
    SparkDFExecutionEngine,
    SqlAlchemyExecutionEngine,
)
from great_expectations.expectations.expectation import (
    ColumnMapExpectation,
    ExpectationValidationResult,
    render_evaluation_parameter_string,
)
from great_expectations.expectations.metrics import (
    ColumnMapMetricProvider,
    column_condition_partial,
)
from great_expectations.expectations.metrics.import_manager import F
from great_expectations.expectations.metrics.metric_provider import metric_partial
from great_expectations.render import CollapseContent, RenderedStringTemplateContent
from great_expectations.render.renderer.renderer import renderer
from great_expectations.validator.metric_configuration import MetricConfiguration


# This class defines a Metric to support your Expectation.
# For most ColumnMapExpectations, the main business logic for calculation will live in this class.
# <snippet name="tests/integration/docusaurus/expectations/creating_custom_expectations/expect_column_values_to_equal_three.py ColumnValuesEqualThree class_def">
class ColumnValuesEqualThree(ColumnMapMetricProvider):
    # </snippet>

    # This is the id string that will be used to reference your metric.
    # <snippet name="tests/integration/docusaurus/expectations/creating_custom_expectations/expect_column_values_to_equal_three.py metric_name">
    condition_metric_name = "column_values.equal_three"
    # </snippet>

    # This method implements the core logic for the PandasExecutionEngine
    # <snippet name="tests/integration/docusaurus/expectations/creating_custom_expectations/expect_column_values_to_equal_three.py pandas">
    @column_condition_partial(engine=PandasExecutionEngine)
    def _pandas(cls, column, **kwargs):
        return column == 3
        # </snippet>

    # <snippet name="tests/integration/docusaurus/expectations/creating_custom_expectations/expect_column_values_to_equal_three.py spark_definition">
    @metric_partial(
        engine=SparkDFExecutionEngine,
        partial_fn_type=MetricPartialFunctionTypes.MAP_CONDITION_FN,
        domain_type=MetricDomainTypes.COLUMN,
    )
    def _spark(
        cls,
        execution_engine: SparkDFExecutionEngine,
        metric_domain_kwargs,
        metric_value_kwargs,
        metrics,
        runtime_configuration,
    ):
        # </snippet>
        # <snippet name="tests/integration/docusaurus/expectations/creating_custom_expectations/expect_column_values_to_equal_three.py spark_selectable">
        (
            selectable,
            compute_domain_kwargs,
            accessor_domain_kwargs,
        ) = execution_engine.get_compute_domain(
            metric_domain_kwargs, MetricDomainTypes.COLUMN
        )

        column_name = accessor_domain_kwargs["column"]
        column = F.col(column_name)
        # </snippet>
        # <snippet name="tests/integration/docusaurus/expectations/creating_custom_expectations/expect_column_values_to_equal_three.py spark_query">
        query = F.when(column == 3, F.lit(False)).otherwise(F.lit(True))

        return (query, compute_domain_kwargs, accessor_domain_kwargs)
        # </snippet>

    # <snippet name="tests/integration/docusaurus/expectations/creating_custom_expectations/expect_column_values_to_equal_three.py sqlalchemy">
    @column_condition_partial(engine=SqlAlchemyExecutionEngine)
    def _sqlalchemy(cls, column, **kwargs):
        return column.in_([3])
        # </snippet>

    @classmethod
    def _get_evaluation_dependencies(
        cls,
        metric: MetricConfiguration,
        configuration: Optional[ExpectationConfiguration] = None,
        execution_engine: Optional[ExecutionEngine] = None,
        runtime_configuration: Optional[Dict] = None,
    ):
        """Returns a dictionary of given metric names and their corresponding configuration, specifying the metric
        types and their respective domains"""
        dependencies: Dict = super()._get_evaluation_dependencies(
            metric=metric,
            configuration=configuration,
            execution_engine=execution_engine,
            runtime_configuration=runtime_configuration,
        )

        table_domain_kwargs: Dict = {
            k: v for k, v in metric.metric_domain_kwargs.items() if k != "column"
        }
        dependencies["table.column_types"] = MetricConfiguration(
            metric_name="table.column_types",
            metric_domain_kwargs=table_domain_kwargs,
            metric_value_kwargs={
                "include_nested": True,
            },
        )

        return dependencies


# This class defines the Expectation itself
# <snippet name="tests/integration/docusaurus/expectations/creating_custom_expectations/expect_column_values_to_equal_three.py ExpectColumnValuesToEqualThree class_def">
class ExpectColumnValuesToEqualThree(ColumnMapExpectation):
    # </snippet>
    # <snippet name="tests/integration/docusaurus/expectations/creating_custom_expectations/expect_column_values_to_equal_three.py docstring">
    """Expect values in this column to equal 3."""
    # </snippet>
    # These examples will be shown in the public gallery.
    # They will also be executed as unit tests for your Expectation.
    # <snippet name="tests/integration/docusaurus/expectations/creating_custom_expectations/expect_column_values_to_equal_three.py examples">
    examples = [
        {
            "data": {
                "all_threes": [3, 3, 3, 3, 3],
                "some_zeroes": [3, 3, 3, 0, 0],
            },
            "tests": [
                {
                    "title": "basic_positive_test",
                    "exact_match_out": False,
                    "include_in_gallery": True,
                    "in": {"column": "all_threes"},
                    "out": {
                        "success": True,
                    },
                },
                {
                    "title": "basic_negative_test",
                    "exact_match_out": False,
                    "include_in_gallery": True,
                    "in": {"column": "some_zeroes", "mostly": 0.8},
                    "out": {
                        "success": False,
                    },
                },
            ],
        }
    ]
    # </snippet>

    # This is the id string of the Metric used by this Expectation.
    # For most Expectations, it will be the same as the `condition_metric_name` defined in your Metric class above.
    # <snippet name="tests/integration/docusaurus/expectations/creating_custom_expectations/expect_column_values_to_equal_three.py map_metric">
    map_metric = "column_values.equal_three"
    # </snippet>

    # This is a list of parameter names that can affect whether the Expectation evaluates to True or False
    # Please see https://docs.greatexpectations.io/en/latest/reference/core_concepts/expectations/expectations.html#expectation-concepts-domain-and-success-keys
    # for more information about domain and success keys, and other arguments to Expectations
    success_keys = ("mostly",)

    # This dictionary contains default values for any parameters that should have default values
    default_kwarg_values = {}

    @renderer(renderer_type="renderer.diagnostic.observed_value")
    @render_evaluation_parameter_string
    def _diagnostic_observed_value_renderer(
        cls,
        configuration: ExpectationConfiguration = None,
        result: ExpectationValidationResult = None,
        runtime_configuration: dict = None,
        **kwargs,
    ):
        assert result, "Must provide a result object."

        result_dict = result.result
        if result_dict is None:
            return "--"

        if result_dict.get("observed_value"):
            observed_value = result_dict.get("observed_value")
            if isinstance(observed_value, (int, float)) and not isinstance(
                observed_value, bool
            ):
                return num_to_str(observed_value, precision=10, use_locale=True)
            return str(observed_value)
        elif result_dict.get("unexpected_percent") is not None:
            return (
                num_to_str(result_dict.get("unexpected_percent"), precision=5)
                + "% unexpected"
            )
        else:
            return "--"

    @renderer(renderer_type="renderer.diagnostic.unexpected_statement")
    @render_evaluation_parameter_string
    def _diagnostic_unexpected_statement_renderer(
        cls,
        configuration: ExpectationConfiguration = None,
        result: ExpectationValidationResult = None,
        runtime_configuration: dict = None,
        **kwargs,
    ):
        assert result, "Must provide a result object."

        success = result.success
        result = result.result

        if result.exception_info["raised_exception"]:
            exception_message_template_str = (
                "\n\n$expectation_type raised an exception:\n$exception_message"
            )

            exception_message = RenderedStringTemplateContent(
                **{
                    "content_block_type": "string_template",
                    "string_template": {
                        "template": exception_message_template_str,
                        "params": {
                            "expectation_type": result.expectation_config.expectation_type,
                            "exception_message": result.exception_info[
                                "exception_message"
                            ],
                        },
                        "tag": "strong",
                        "styling": {
                            "classes": ["text-danger"],
                            "params": {
                                "exception_message": {"tag": "code"},
                                "expectation_type": {
                                    "classes": ["badge", "badge-danger", "mb-2"]
                                },
                            },
                        },
                    },
                }
            )

            exception_traceback_collapse = CollapseContent(
                **{
                    "collapse_toggle_link": "Show exception traceback...",
                    "collapse": [
                        RenderedStringTemplateContent(
                            **{
                                "content_block_type": "string_template",
                                "string_template": {
                                    "template": result.exception_info[
                                        "exception_traceback"
                                    ],
                                    "tag": "code",
                                },
                            }
                        )
                    ],
                }
            )

            return [exception_message, exception_traceback_collapse]

        if success or not result_dict.get("unexpected_count"):
            return []
        else:
            unexpected_count = num_to_str(
                result_dict["unexpected_count"], use_locale=True, precision=20
            )
            unexpected_percent = (
                num_to_str(result_dict["unexpected_percent"], precision=4) + "%"
            )
            element_count = num_to_str(
                result_dict["element_count"], use_locale=True, precision=20
            )

            template_str = (
                "\n\n$unexpected_count unexpected values found. "
                "$unexpected_percent of $element_count total rows."
            )

            return [
                RenderedStringTemplateContent(
                    **{
                        "content_block_type": "string_template",
                        "string_template": {
                            "template": template_str,
                            "params": {
                                "unexpected_count": unexpected_count,
                                "unexpected_percent": unexpected_percent,
                                "element_count": element_count,
                            },
                            "tag": "strong",
                            "styling": {"classes": ["text-danger"]},
                        },
                    }
                )
            ]

    @renderer(renderer_type="renderer.diagnostic.unexpected_table")
    @render_evaluation_parameter_string
    def _diagnostic_unexpected_table_renderer(
        cls,
        configuration: ExpectationConfiguration = None,
        result: ExpectationValidationResult = None,
        runtime_configuration: dict = None,
        **kwargs,
    ):
        try:
            result_dict = result.result
        except KeyError:
            return None

        if result_dict is None:
            return None

        if not result_dict.get("partial_unexpected_list") and not result_dict.get(
            "partial_unexpected_counts"
        ):
            return None

        table_rows = []

        if result_dict.get("partial_unexpected_counts"):
            total_count = 0
            for unexpected_count_dict in result_dict.get("partial_unexpected_counts"):
                value = unexpected_count_dict.get("value")
                count = unexpected_count_dict.get("count")
                total_count += count
                if value is not None and value != "":
                    table_rows.append([value, count])
                elif value == "":
                    table_rows.append(["EMPTY", count])
                else:
                    table_rows.append(["null", count])

            if total_count == result_dict.get("unexpected_count"):
                header_row = ["Unexpected Value", "Count"]
            else:
                header_row = ["Sampled Unexpected Values"]
                table_rows = [[row[0]] for row in table_rows]

        else:
            header_row = ["Sampled Unexpected Values"]
            sampled_values_set = set()
            for unexpected_value in result_dict.get("partial_unexpected_list"):
                if unexpected_value:
                    string_unexpected_value = str(unexpected_value)
                elif unexpected_value == "":
                    string_unexpected_value = "EMPTY"
                else:
                    string_unexpected_value = "null"
                if string_unexpected_value not in sampled_values_set:
                    table_rows.append([unexpected_value])
                    sampled_values_set.add(string_unexpected_value)

        unexpected_table_content_block = RenderedTableContent(
            **{
                "content_block_type": "table",
                "table": table_rows,
                "header_row": header_row,
                "styling": {
                    "body": {"classes": ["table-bordered", "table-sm", "mt-3"]}
                },
            }
        )

        return unexpected_table_content_block

    # This dictionary contains metadata for display in the public gallery
    # <snippet name="tests/integration/docusaurus/expectations/creating_custom_expectations/expect_column_values_to_equal_three.py library_metadata">
    library_metadata = {
        "tags": ["extremely basic math"],
        "contributors": ["@joegargery"],
    }
    # </snippet>


if __name__ == "__main__":
    # <snippet name="tests/integration/docusaurus/expectations/creating_custom_expectations/expect_column_values_to_equal_three.py diagnostics">
    ExpectColumnValuesToEqualThree().print_diagnostic_checklist()
#     </snippet>

# Note to users: code below this line is only for integration testing -- ignore!

diagnostics = ExpectColumnValuesToEqualThree().run_diagnostics()

for check in diagnostics["tests"]:
    assert check["test_passed"] is True
    assert check["error_diagnostics"] is None

for check in diagnostics["errors"]:
    assert check is None

for check in diagnostics["maturity_checklist"]["experimental"]:
    if check["message"] == "Passes all linting checks":
        continue
    assert check["passed"] is True



from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Dict, Optional

import numpy as np
from typing_extensions import Final

from great_expectations.core.domain import Domain  # noqa: TCH001
from great_expectations.rule_based_profiler.estimators.numeric_range_estimation_result import (
    NumericRangeEstimationResult,  # noqa: TCH001
)
from great_expectations.rule_based_profiler.estimators.numeric_range_estimator import (
    NumericRangeEstimator,
)
from great_expectations.rule_based_profiler.helpers.util import (
    compute_quantiles,
    datetime_semantic_domain_type,
    get_false_positive_rate_from_rule_state,
    get_quantile_statistic_interpolation_method_from_rule_state,
)
from great_expectations.rule_based_profiler.parameter_container import (
    ParameterContainer,  # noqa: TCH001
)
from great_expectations.types.attributes import Attributes  # noqa: TCH001
from great_expectations.util import convert_ndarray_to_datetime_dtype_best_effort

if TYPE_CHECKING:
    import numpy.typing as npt

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

DEFAULT_QUANTILES_QUANTILE_STATISTIC_INTERPOLATION_METHOD: Final[str] = "nearest"


class QuantilesNumericRangeEstimator(NumericRangeEstimator):
    """
    Implements "quantiles" computation.

    This nonparameteric estimator calculates quantiles given a MetricValues vector of length N, the q-th quantile of
        the vector is the value q of the way from the minimum to the maximum in a sorted copy of the MetricValues.
    """

    def __init__(
        self,
        configuration: Optional[Attributes] = None,
    ) -> None:
        super().__init__(
            name="quantiles",
            configuration=configuration,
        )

    def _get_numeric_range_estimate(
        self,
        metric_values: np.ndarray,
        domain: Domain,
        variables: Optional[ParameterContainer] = None,
        parameters: Optional[Dict[str, ParameterContainer]] = None,
    ) -> NumericRangeEstimationResult:
        false_positive_rate: np.float64 = get_false_positive_rate_from_rule_state(  # type: ignore[assignment] # could be float
            false_positive_rate=self.configuration.false_positive_rate,  # type: ignore[union-attr] # configuration could be None
            domain=domain,
            variables=variables,
            parameters=parameters,
        )
        quantile_statistic_interpolation_method: str = get_quantile_statistic_interpolation_method_from_rule_state(
            quantile_statistic_interpolation_method=self.configuration.quantile_statistic_interpolation_method,  # type: ignore[union-attr] # configuration could be None
            round_decimals=self.configuration.round_decimals,  # type: ignore[union-attr] # configuration could be None
            domain=domain,
            variables=variables,
            parameters=parameters,
        )
        if quantile_statistic_interpolation_method is None:
            quantile_statistic_interpolation_method = (
                DEFAULT_QUANTILES_QUANTILE_STATISTIC_INTERPOLATION_METHOD
            )

        datetime_detected: bool = datetime_semantic_domain_type(domain=domain)
        metric_values_converted: npt.NDArray
        (
            _,
            _,
            metric_values_converted,
        ) = convert_ndarray_to_datetime_dtype_best_effort(
            data=metric_values,
            datetime_detected=datetime_detected,
            parse_strings_as_datetimes=True,
            fuzzy=False,
        )
        return compute_quantiles(
            metric_values=metric_values_converted,
            false_positive_rate=false_positive_rate,
            quantile_statistic_interpolation_method=quantile_statistic_interpolation_method,
        )



import logging
from dataclasses import asdict, dataclass
from enum import Enum

from great_expectations.core.util import convert_to_json_serializable, nested_update
from great_expectations.rule_based_profiler.helpers.util import (
    convert_variables_to_dict,
)
from great_expectations.rule_based_profiler.parameter_container import (
    ParameterContainer,  # noqa: TCH001
)
from great_expectations.types import SerializableDictDot

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


class ReconciliationStrategy(Enum):
    NESTED_UPDATE = "nested_update"
    REPLACE = "replace"
    UPDATE = "update"


@dataclass
class ReconciliationDirectives(SerializableDictDot):
    variables: ReconciliationStrategy = ReconciliationStrategy.UPDATE
    domain_builder: ReconciliationStrategy = ReconciliationStrategy.UPDATE
    parameter_builder: ReconciliationStrategy = ReconciliationStrategy.UPDATE
    expectation_configuration_builder: ReconciliationStrategy = (
        ReconciliationStrategy.UPDATE
    )

    def to_dict(self) -> dict:
        return asdict(self)

    def to_json_dict(self) -> dict:
        return convert_to_json_serializable(data=self.to_dict())


DEFAULT_RECONCILATION_DIRECTIVES = ReconciliationDirectives(
    variables=ReconciliationStrategy.UPDATE,
    domain_builder=ReconciliationStrategy.UPDATE,
    parameter_builder=ReconciliationStrategy.UPDATE,
    expectation_configuration_builder=ReconciliationStrategy.UPDATE,
)


def reconcile_rule_variables(
    variables: ParameterContainer,
    variables_config: dict,
    reconciliation_strategy: ReconciliationStrategy = DEFAULT_RECONCILATION_DIRECTIVES.variables,
) -> dict:
    """
    Rule "variables" reconciliation involves combining the variables, instantiated from Rule configuration
    (e.g., stored in a YAML file managed by the Profiler store), with the variables override, possibly supplied
    as part of the candiate override rule configuration.

    The reconciliation logic for "variables" is of the "replace" nature: An override value complements the
    original on key "miss", and replaces the original on key "hit" (or "collision"), because "variables" is a
    unique member for a Rule.

    :param variables: existing variables of a Rule
    :param variables_config: variables configuration override, supplied in dictionary (configuration) form
    :param reconciliation_strategy: one of update, nested_update, or overwrite ways of reconciling overwrites
    :return: reconciled variables configuration, returned in dictionary (configuration) form
    """
    effective_variables_config: dict = convert_variables_to_dict(variables=variables)
    if variables_config:
        if reconciliation_strategy == ReconciliationStrategy.NESTED_UPDATE:
            effective_variables_config = nested_update(
                effective_variables_config,
                variables_config,
            )
        elif reconciliation_strategy == ReconciliationStrategy.REPLACE:
            effective_variables_config = variables_config
        elif reconciliation_strategy == ReconciliationStrategy.UPDATE:
            effective_variables_config.update(variables_config)

    return effective_variables_config



from __future__ import annotations

import io
from pathlib import Path

from ruamel.yaml import YAML

from great_expectations.alias_types import JSONValues  # noqa: TCH001
from great_expectations.core._docs_decorators import public_api


@public_api
class YAMLHandler:
    """Facade class designed to be a lightweight wrapper around YAML serialization.

    For all YAML-related activities in Great Expectations, this is the entry point.

    Note that this is meant to be library agnostic - the underlying implementation does not
    matter as long as we fulfill the following contract:

    * load
    * dump

    Typical usage example:

    ```python
    simple_yaml: str = '''
        name: test
        class_name: test_class
        module_name: test.test_class
    '''
    yaml_handler = YAMLHandler()
    res: dict = yaml_handler.load(simple_yaml)
    example_dict: dict = dict(abc=1)
    yaml_handler.dump(example_dict)
    ```

    """

    def __init__(self) -> None:
        self._handler = YAML(typ="safe")
        # TODO: ensure this does not break all usage of ruamel in GX codebase.
        self._handler.indent(mapping=2, sequence=4, offset=2)
        self._handler.default_flow_style = False

    @public_api
    def load(self, stream: io.TextIOWrapper | str) -> dict[str, JSONValues]:
        """Converts a YAML input stream into a Python dictionary.

        Example:

        ```python
        import pathlib
        yaml_handler = YAMLHandler()
        my_file_str = pathlib.Path("my_file.yaml").read_text()
        dict_from_yaml = yaml_handler.load(my_file_str)
        ```

        Args:
            stream: The input stream to read in. Although this function calls ruamel's load(), we
                use a slightly more restrictive type-hint than ruamel (which uses Any). This is in order to tightly
                bind the behavior of the YamlHandler class with expected YAML-related activities of Great Expectations.

        Returns:
            The deserialized dictionary form of the input stream.
        """
        return self._handler.load(stream=stream)

    @public_api
    def dump(
        self,
        data: dict,
        stream: io.TextIOWrapper | io.StringIO | Path | None = None,
        **kwargs,
    ) -> str | None:
        """Converts a Python dictionary into a YAML string.

        Dump code has been adopted from:
        https://yaml.readthedocs.io/en/latest/example.html#output-of-dump-as-a-string

        ```python
        >>> data = {'foo': 'bar'}
        >>> yaml_str = yaml_handler.dump(data)
        >>> print(yaml_str)
        foo:
            bar:
        ```

        Args:
            data: The dictionary to serialize into a Python object.
            stream: The output stream to modify. If not provided, we default to io.StringIO.
            kwargs: Additional key-word arguments to pass to underlying yaml dump method.

        Returns:
            If no stream argument is provided, the str that results from ``_handler.dump()``.
            Otherwise, None as the ``_handler.dump()`` works in place and will exercise the handler accordingly.
        """
        if stream:
            return self._dump(data=data, stream=stream, **kwargs)  # type: ignore[func-returns-value]
        return self._dump_and_return_value(data=data, **kwargs)

    def _dump(self, data: dict, stream, **kwargs) -> None:
        """If an input stream has been provided, modify it in place."""
        self._handler.dump(data=data, stream=stream, **kwargs)

    def _dump_and_return_value(self, data: dict, **kwargs) -> str:
        """If an input stream hasn't been provided, generate one and return the value."""
        stream = io.StringIO()
        self._handler.dump(data=data, stream=stream, **kwargs)
        return stream.getvalue()



from abc import ABC, abstractmethod


class BaseUpgradeHelper(ABC):
    """Base UpgradeHelper abstract class"""

    @abstractmethod
    def manual_steps_required(self) -> None:
        pass

    @abstractmethod
    def get_upgrade_overview(self) -> None:
        pass

    @abstractmethod
    def upgrade_project(self) -> None:
        pass

    @abstractmethod
    def _generate_upgrade_report(self) -> None:
        pass

    @abstractmethod
    def _save_upgrade_log(self) -> None:
        pass
